import logging
from copy import deepcopy
from typing import Dict, List

from nicegui import ui

import ieee_2030_5.models as m

from ..pages import Pages, show_global_header
from ..session import get_control_list, get_curve_list, send_control

_log = logging.getLogger(__name__)

# Curves and Curve Type are here only for read only purposes.
curve_list: m.DERCurveList = m.DERCurveList()
curve_type_filtered: Dict[m.CurveType, List[m.DERCurve]] = {}

# This is the list of controls that are available.
controls = m.DERControlList()

# The currrent control binding for this form.
current_control = m.DERControl()
current_control.opModConnect = False
current_control.opModEnergize = False

# The current dercontrol and dercontrolbase are here for read/write purposes.
der_base = m.DERControlBase()
der_default = m.DefaultDERControl()

der_base.DERControlBase = der_default
current_control.DERControlBase = der_base


@ui.refreshable
def render_select():
    der_controls = [c.description for c in controls.DERControl]
    der_controls.insert(0, "NEW")
    with ui.row():
        with ui.column():
            if not current_control.description:
                ui.select(der_controls, label="Controls", value=der_controls[0]).classes("w-64")
            else:
                ui.select(der_controls, label="Controls",
                          value=current_control.description).classes("w-64")

            if current_control.href:
                ui.label(f"Href: {current_control.href}")


def change_control(new_control: str):
    pass


@ui.refreshable
def render_der_control_form():
    with ui.row():
        ui.input("description").classes("w-64") \
            .bind_value(current_control, "description")
    ui.separator()

    with ui.row():
        with ui.column().classes("pr-20"):
            ui.label("DERDefaultControl")

            esdelay_input = ui.input("setESDelay (hundredth of a second)") \
                .bind_value(der_default, "setESDelay").props('clearable')
            setESHighFreq = ui.input("setESHighFreq (hundredth of a hertz)") \
                .bind_value(der_default, "setESHighFreq")
            setESHighVolt = ui.input("setESHighVolt (hundredth of a volt)") \
                .bind_value(der_default, "setESHighVolt")
            setESLowFreq = ui.input("setESLowFreq (hundredth of a hertz)") \
                .bind_value(der_default, "setESLowFreq")
            setESLowVolt = ui.input("setESLowVolt (hundredth of a volt)") \
                .bind_value(der_default, "setESLowVolt")
            setESRampTms = ui.input("setESRampTms (hundredth of a second)") \
                .bind_value(der_default, "setESRampTms")
            setESRandomDelay = ui.input("setESRandomDelay (hundredth of a second)") \
                .bind_value(der_default, "setESRandomDelay")
            setGradW = ui.input("setGradW (hundredth of a watt)") \
                .bind_value(der_default, "setGradW")
            setSoftGradW = ui.input("setSoftGradW (hundredth of a watt)") \
                .bind_value(der_default, "setSoftGradW")

        with ui.column().classes("pr-20"):
            ui.label("DERControlBase")

            opModConnect = ui.checkbox("opModConnect") \
                .bind_value(der_base, "opModConnect")
            opModEnergize = ui.checkbox("opModEnergize") \
                .bind_value(der_base, "opModEnergize")
            opModFixedPFAbsorbW = ui.input("opModFixedPFAbsorbW") \
                .bind_value(der_base, "opModFixedPFAbsorbW")
            opModFixedPFAbsorbW = ui.input("opModFixedPFAbsorbW") \
                .bind_value(der_base, "opModFixedPFAbsorbW")

            opModFixedPFInjectW = ui.input("opModFixedPFInjectW") \
                .bind_value(der_base, "opModFixedPFInjectW")

            opModFixedVar = ui.input("opModFixedVar") \
                .bind_value(der_base, "opModFixedVar")
            opModFixedW = ui.input("opModFixedW") \
                .bind_value(der_base, "opModFixedW")
            opModFreqDroop = ui.input("opModFreqDroop") \
                .bind_value(der_base, "opModFreqDroop")


            opModMaxLimW = ui.input("opModMaxLimW") \
                .bind_value(der_base, "opModMaxLimW")
            opModTargetVar = ui.input("opModTargetVar") \
                .bind_value(der_base, "opModTargetVar")
            opModTargetW = ui.input("opModTargetW") \
                .bind_value(der_base, "opModTargetW")
            opModVoltVar = ui.input("opModVoltVar") \
                .bind_value(der_base, "opModVoltVar")
            opModWattPF = ui.input("opModWattPF") \
                .bind_value(der_base, "opModWattPF")

            rampTms: ui.input("rampTms") \
                .bind_value(der_base, "rampTms")

        with ui.column().classes("pr-20"):
            ui.label("Curve Selection")

            pModFreqWatt = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModFreqWatt]}, label="pModFreqWatt").classes("w-64") \
                .bind_value_from(der_base, "pModFreqWatt")
            pModFreqWatt.enabled = len(curve_type_filtered[m.CurveType.opModFreqWatt]) > 0


            opModHFRTMayTrip = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModHFRTMayTrip]}, label="opModHFRTMayTrip").classes("w-64")\
                .bind_value_from(der_base, "opModHFRTMayTrip")
            opModHFRTMayTrip.enabled = len(curve_type_filtered[m.CurveType.opModHFRTMayTrip]) > 0


            opModHFRTMustTrip = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModHFRTMustTrip]}, label="opModHFRTMustTrip").classes("w-64")\
                .bind_value_from(der_base, "opModHFRTMustTrip")
            opModHFRTMustTrip.enabled = len(curve_type_filtered[m.CurveType.opModHFRTMustTrip]) > 0

            opModHVRTMayTrip = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModHVRTMayTrip]}, label="opModHVRTMayTrip").classes("w-64") \
                .bind_value_from(der_base, "opModHVRTMayTrip")
            opModHVRTMayTrip.enabled = len(curve_type_filtered[m.CurveType.opModHVRTMayTrip]) > 0

            opModHVRTMustTrip = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModHVRTMustTrip]}, label="opModHVRTMustTrip").classes("w-64") \
                .bind_value_from(der_base, "opModHVRTMustTrip")
            opModHVRTMustTrip.enabled = len(curve_type_filtered[m.CurveType.opModHFRTMustTrip]) > 0

            opModHVRTMomentaryCessation = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModHVRTMomentaryCessation]}, label="opModHVRTMomentaryCessation").classes("w-64") \
                .bind_value_from(der_base, "opModHVRTMomentaryCessation")
            opModHVRTMomentaryCessation.enabled = len(
                curve_type_filtered[m.CurveType.opModHVRTMomentaryCessation]) > 0

            opModLFRTMayTrip = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModLFRTMayTrip]}, label="opModLFRTMayTrip").classes("w-64") \
                .bind_value_from(der_base, "opModLFRTMayTrip")
            opModLFRTMayTrip.enabled = len(curve_type_filtered[m.CurveType.opModLFRTMayTrip]) > 0

            opModLVRTMayTrip = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModLVRTMayTrip]}, label="opModLVRTMayTrip").classes("w-64") \
                .bind_value_from(der_base, "opModLVRTMayTrip")
            opModLVRTMayTrip.enabled = len(curve_type_filtered[m.CurveType.opModLVRTMayTrip]) > 0

            opModLVRTMustTrip = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModLVRTMustTrip]}, label="opModLVRTMustTrip").classes("w-64") \
                .bind_value_from(der_base, "opModLVRTMustTrip")
            opModLVRTMustTrip.enabled = len(curve_type_filtered[m.CurveType.opModLVRTMustTrip]) > 0

            opModLVRTMomentaryCessation = ui.select({v.href: v.description for v in curve_type_filtered[m.CurveType.opModLVRTMomentaryCessation]}, label="opModLVRTMomentaryCessation").classes("w-64") \
                .bind_value_from(der_base, "opModLVRTMomentaryCessation")
            opModLVRTMomentaryCessation.enabled = len(
                curve_type_filtered[m.CurveType.opModLVRTMomentaryCessation]) > 0


def revert_changes():
    pass


def validate_and_submit():
    """Validate the objects and submit them to the server.
    
    Returns:
        bool: True if successful, False otherwise.
    """
    global der_base, der_default, current_control

    if append_new_control := current_control.href is None:
        ...

    current_control.DERControlBase = der_base
    control = send_control(current_control)

    assert isinstance(control, m.DERControl)

    if append_new_control:
        controls.DERControl.append(control)

    current_control = control
    render_select.refresh()
    #render_select.refresh()
    # global curve_data_points, current_curve
    # current_curve.CurveData = curve_data_points

    # if append_new_curve := current_curve.href is None:
    #     ...

    # # POST or PUT the curve to the server
    # dc = send_control(current_curve)

    # assert isinstance(dc, m.DERCurve)

    # # update the variables for state
    # curve_data_points = dc.CurveData
    # current_curve = dc

    # # IF its post then we append to the curve list
    # if append_new_curve:
    #     curve_list.DERCurve.append(dc)

    # render_select.refresh()

    ui.notify("Control saved successfully")


@ui.page("/controls")
def show_controls():
    global curve_list, curve_type_filtered

    curve_response = get_curve_list()
    if curve_response:
        curve_list = curve_response
        curve_type_filtered.clear()
        for curve_type in m.CurveType:
            curve_type_filtered[curve_type] = [
                curve for curve in curve_list.DERCurve
                if curve.curveType == curve_type.name or curve.curveType == curve_type.value
            ]

    show_global_header(Pages.CONTROLS)
    render_select()
    render_der_control_form()

    ui.separator()
    with ui.row():
        ui.button("Store", on_click=validate_and_submit)
        ui.button("Revert", on_click=revert_changes)


from datetime import datetime

import pytz
from flask import Response, request

import ieee_2030_5.adapters as adpt
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.types_ import format_time
from ieee_2030_5.utils import xml_to_dataclass


class Log(RequestOp):

    def get(self) -> Response:
        pth = request.environ['PATH_INFO']
        return self.build_response_from_dataclass(adpt.LogAdapter.fetch_list(pth))

    def post(self) -> Response:
        """Posting of log event allows client to store information for a display to get.
        
        For 2030.5 this is posted at an end device level so that its available to the
        server.
        """
        path = request.environ['PATH_INFO']
        data: m.LogEvent = xml_to_dataclass(request.data.decode('utf-8'))
        data_type = type(data)
        if data_type not in (m.LogEvent):
            raise BAD_REQUEST()

        if not data.createdDateTime:
            data.createdDateTime = format_time(datetime.utcnow().replace(tzinfo=pytz.utc))
        adpt.LogAdapter.store(path, data)


from enum import IntEnum


class PrimacyType(IntEnum):
    # Values possible for indication of "Primary" provider:
    # 0: In home energy management system
    # 1: Contracted premises service provider
    # 2: Non-contractual service provider
    # 3 - 64: Reserved
    # 65 - 191: User-defined
    #192 - 255: Reserved
    InHomeManagementSystem = 0
    ContractedPremisesServiceProvider = 1
    NonContractualServiceProvider = 2
        

class DERUnitRefType(IntEnum):
    # 0 - N/A
    # 1 - %setMaxW
    # 2 - %setMaxVar
    # 3 - %statVarAvail
    # 4 - %setEffectiveV
    # 5 - %setMaxChargeRateW
    # 6 - %setMaxDischargeRateW
    # 7 - %statWAvail
    NA = 0
    setMaxW = 1
    setMaxVar = 2
    statVarAvail = 3
    setEffectiveV = 4
    setMaxChargeRateW = 5
    setMaxDischargeRateW = 6
    statWAvail = 7

class CurveType(IntEnum):
    # 0 - opModFreqWatt (Frequency-Watt Curve Mode)
    # 1 - opModHFRTMayTrip (High Frequency Ride Through, May Trip Mode)
    # 2 - opModHFRTMustTrip (High Frequency Ride Through, Must Trip Mode)
    # 3 - opModHVRTMayTrip (High Voltage Ride Through, May Trip Mode)
    # 4 - opModHVRTMomentaryCessation (High Voltage Ride Through, Momentary Cessation
    # Mode)
    # 5 - opModHVRTMustTrip (High Voltage Ride Through, Must Trip Mode)
    # 6 - opModLFRTMayTrip (Low Frequency Ride Through, May Trip Mode)
    # 7 - opModLFRTMustTrip (Low Frequency Ride Through, Must Trip Mode)
    # 8 - opModLVRTMayTrip (Low Voltage Ride Through, May Trip Mode)
    # 9 - opModLVRTMomentaryCessation (Low Voltage Ride Through, Momentary Cessation
    # Mode)
    # 10 - opModLVRTMustTrip (Low Voltage Ride Through, Must Trip Mode)
    # 11 - opModVoltVar (Volt-Var Mode)
    # 12 - opModVoltWatt (Volt-Watt Mode)
    # 13 - opModWattPF (Watt-PowerFactor Mode)
    # 14 - opModWattVar (Watt-Var Mode)
    opModFreqWatt = 0
    opModHFRTMayTrip = 1
    opModHFRTMustTrip = 2
    opModHVRTMayTrip = 3
    opModHVRTMomentaryCessation = 4
    opModHVRTMustTrip = 5
    opModLFRTMayTrip = 6
    opModLFRTMustTrip = 7
    opModLVRTMayTrip = 8
    opModLVRTMomentaryCessation = 9
    opModLVRTMustTrip = 10
    opModVoltVar = 11
    opModVoltWatt = 12
    opModWattPF = 13
    opModWattVar = 14


class DeviceCategoryType(IntEnum):
    """
    DeviceCategoryType defined from 20305-2018_IIEStandardforSmartEnergyProfileApplicationsProtocol.pdf Appendix
    B.2.3.4 Types package
    """
    # The Device category types defined.
    # Bit positions SHALL be defined as follows:
    PROGRAMMABLE_COMMUNICATING_THERMOSTAT = 0
    STRIP_HEATERS = 1
    BASEBOARD_HEATERS = 2
    WATER_HEATER = 3
    POOL_PUMP = 4
    SAUNA = 5
    HOT_TUB = 6
    SMART_APPLIANCE = 7
    IRRIGATION_PUMP = 8
    MANAGED_COMMERCIAL_AND_INDUSTRIAL_LOADS = 9
    SIMPLE_RESIDENTIAL_LOADS = 10    # On/Off loads
    EXTERIOR_LIGHTING = 11
    INTERIOR_LIGHTING = 12
    LOAD_CONTROL_SWITCH = 13
    ENERGY_MANAGEMENT_SYSTEM = 14
    SMART_ENERGY_MODULE = 15
    ELECTRIC_VEHICLE = 16
    ELECTRIC_VEHICLE_SUPPLY_EQUIPMENT = 17
    VIRTUAL_OR_MIXED_DER = 18
    RECIPROCATING_ENGINE = 19    # Synchronous Machine
    FUEL_CELL = 20    # Battery
    PHOTOVOLTAIC_SYSTEM = 21    # Solar
    COMBINED_HEAT_AND_POWER = 22
    COMBINED_PV_AND_STORAGE = 23
    OTHER_GENERATION_SYSTEMS = 24
    OTHER_STORAGE_SYSTEMS = 25

    # Additional here for Aggregator
    AGGREGATOR = 99
    OTHER_CLIENT = 100


#     0 - Programmable Communicating Thermostat
#     1 - Strip Heaters
#     2 - Baseboard Heaters
#     3 - Water Heater
#     4 - Pool Pump
#     5 - Sauna
#     6 - Hot tub
#     7 - Smart Appliance
#     8 - Irrigation Pump
#     9 - Managed Commercial and Industrial Loads
#     10 - Simple Residential Loads
#     11 - Exterior Lighting
#     12 - Interior Lighting
#     13 - Electric Vehicle
#     14 - Generation Systems
#     15 - Load Control Switch
#     16 - Smart Inverter
#     17 - EVSE
#     18 - Residential Energy Storage Unit
#     19 - Energy Management System
#     20 - Smart Energy Module



import json
import logging
from pathlib import Path
from typing import Optional

import flask
from blinker import Signal
from flask import Flask, Response, g, render_template, request

import ieee_2030_5.adapters as adpt
from ieee_2030_5.data.indexer import add_href, get_href
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.server.server_constructs import create_device_capability
from ieee_2030_5.utils import (dataclass_to_xml, get_lfdi_from_cert, get_sfdi_from_lfdi,
                               xml_to_dataclass)

_log = logging.getLogger(__name__)


class AdminEndpoints:

    def __init__(self, app: Flask, tls_repo: TLSRepository, config: ServerConfiguration):
        self.tls_repo = tls_repo
        self.server_config = config

        app.add_url_rule("/admin", view_func=self._admin)
        #app.add_url_rule("/admin/enddevices/<int:index>", view_func=self._admin_enddevices)
        #app.add_url_rule("/admin/enddevices", view_func=self._admin_enddevices)
        #app.add_url_rule("/admin/end-device-list", view_func=self._admin_enddevice_list)
        #app.add_url_rule("/admin/program-lists", view_func=self._admin_der_program_lists)
        #app.add_url_rule("/admin/lfdi", endpoint="admin/lfdi", view_func=self._lfdi_lists)
        #app.add_url_rule("/admin/edev/<int:edev_index>/ders/<int:der_index>/current_derp", view_func=self._admin_der_update_current_derp, methods=['PUT', 'GET'])

        #        app.add_url_rule("/admin/ders/<int:edev_index>", view_func=self._admin_ders)

        # COMPLETE
        app.add_url_rule("/admin/edev/<int:edevid>/fsa/<int:fsaid>/derp",
                         view_func=self._admin_edev_fsa_derp)
        app.add_url_rule("/admin/edev/<int:edevid>/fsa/<int:fsaid>",
                         view_func=self._admin_edev_fsa)
        app.add_url_rule("/admin/edev/<int:edevid>/fsa", view_func=self._admin_edev_fsa)
        app.add_url_rule("/admin/edev/<int:edevid>/der", view_func=self._admin_edev_ders)
        app.add_url_rule("/admin/edev/<int:edevid>/der/<int:derid>/current_derp",
                         view_func=self._admin_edev_ders)
        app.add_url_rule("/admin/edev/<int:edevid>/der/<int:derid>",
                         view_func=self._admin_edev_ders)
        app.add_url_rule("/admin/edev", view_func=self._admin_edev)
        # END COMPLETE

        app.add_url_rule("/admin/certs", endpoint="admin/certs", view_func=self._admin_certs)
        app.add_url_rule("/admin/enddevices",
                         methods=['GET', 'PUT', 'POST', 'DELETE'],
                         view_func=self._admin_enddevices)
        app.add_url_rule("/admin/curves",
                         methods=['GET', 'PUT', 'POST', 'DELETE'],
                         view_func=self._admin_curves)
        app.add_url_rule("/admin/controls",
                         methods=['GET', 'PUT', 'POST', 'DELETE'],
                         view_func=self._admin_controls)
        app.add_url_rule("/admin/programs",
                         methods=['GET', 'PUT', 'POST', 'DELETE'],
                         view_func=self._admin_programs)
        app.add_url_rule("/admin/fsa",
                         methods=['GET', 'PUT', 'POST', 'DELETE'],
                         view_func=self._admin_fsa)
        app.add_url_rule("/admin/der",
                         methods=['GET', 'PUT', 'POST', 'DELETE'],
                         view_func=self._admin_ders)

        app.add_url_rule("/admin/derp/<int:derp_index>/derc/<int:control_index>",
                         methods=['GET', 'PUT'],
                         view_func=self._admin_derp_derc)
        app.add_url_rule("/admin/derp/<int:derp_index>/derc",
                         methods=['GET', 'POST'],
                         view_func=self._admin_derp_derc)
        app.add_url_rule("/admin/derp/<int:derp_index>/derca",
                         methods=['GET'],
                         view_func=self._admin_derp_derca)
        app.add_url_rule("/admin/derp/<int:derp_index>/dderc",
                         methods=['GET', 'PUT'],
                         view_func=self._admin_derp_derc)
        app.add_url_rule("/admin/derp", methods=['GET', 'POST'], view_func=self._admin_derp)
        #app.add_url_rule("/admin/derp/<int:index>",  methods=['GET', 'POST'], view_func=self._derp)
        #app.add_url_rule("/admin/derp/<int:index>/derc", methods=['GET', 'POST'], view_func=self._derp_derc)

    # def _admin_edev_fsa(self, edevid: int, fsaid: int = -1, derc: int = -1) -> Response:
    #     #edev = EndDeviceAdapter.fetch_at(edevid)

    #     if fsaid > -1 and derc > -1:

    #fsa = FSAAdapter.fetch_by_end_device(edevid)

    def _admin_certs(self) -> Response:
        headers = {'CONTENT-TYPE': "application/json"}
        tls_repo: TLSRepository = g.TLS_REPOSITORY

        return Response(json.dumps(tls_repo.client_list), headers=headers)

    def _admin_ders(self) -> Response:
        """Returns DER or DERList depending on the request method.
        
        If request method is GET, return an DERList object.
        
        If request method is POST, save the DER object and return it with new href for the end device.
        
        If request method is PUT, update the DER object and return it.
        
        If request method is DELETE, remove the DER object and return None.
        """
        if request.method in ('POST', 'PUT'):
            data = request.data.decode('utf-8')
            item = xml_to_dataclass(data)
            if not isinstance(item, m.DER):
                _log.error("DER was not passed via data.")
                return Response(status=400)

            if request.method == 'POST':
                if item.href:
                    _log.error(f"POST method with existing object {item.href}")
                    return Response(400)

                item = adpt.DERAdapter.add(item)
                response_status = 201

            elif request.method == 'PUT':
                if not item.href:
                    _log.error(f"PUT method without an existing object.")
                    return Response(400)

                index = int(item.href.rsplit(hrefs.SEP)[-1])
                adpt.DERAdapter.put(index, item)
                response_status = 200

            return Response(dataclass_to_xml(item), status=response_status)

    def _admin_programs(self) -> Response:
        """Returns EndDevice or EndDeviceList depending on the request method.
        
        If request method is GET, return an EndDeviceList object.
        
        If request method is POST, save the EndDevice object and return it with new href for the end device.
        
        If request method is PUT, update the EndDevice object and return it.
        
        If request method is DELETE, remove the EndDevice object and return None.
        """
        if request.method in ('POST', 'PUT'):

            def normalize_certificate_name(pre_name: str) -> str:
                if pre_name.startswith('/'):
                    pre_name = pre_name[1:]
                return pre_name.replace('/', '-')

            data = request.data.decode('utf-8')
            item = xml_to_dataclass(data)
            if not isinstance(item, m.DERProgram):
                _log.error("DERProgram was not passed via data.")
                return Response(status=400)

            if request.method == 'POST':
                if item.href:
                    _log.error(f"POST method with existing object {item.href}")
                    return Response(400)

                item = adpt.DERProgramAdapter.add(item)
                response_status = 201

            elif request.method == 'PUT':
                if not item.href:
                    _log.error(f"PUT method without an existing object.")
                    return Response(400)

                index = int(item.href.rsplit(hrefs.SEP)[-1])
                adpt.DERProgramAdapter.put(index, item)
                response_status = 200

            return Response(dataclass_to_xml(item), status=response_status)

        # Get all end devices
        start = int(request.args.get('s', 0))
        after = int(request.args.get('a', 0))
        limit = int(request.args.get('l', 1))

        allofem = adpt.DERProgramAdapter.fetch_all(m.DERProgramList(),
                                                   start=start,
                                                   after=after,
                                                   limit=limit)
        return Response(dataclass_to_xml(allofem), status=200)

    def _admin_fsa(self) -> Response:
        """Returns EndDevice or EndDeviceList depending on the request method.
        
        If request method is GET, return an EndDeviceList object.
        
        If request method is POST, save the EndDevice object and return it with new href for the end device.
        
        If request method is PUT, update the EndDevice object and return it.
        
        If request method is DELETE, remove the EndDevice object and return None.
        """
        if request.method in ('POST', 'PUT'):

            def normalize_certificate_name(pre_name: str) -> str:
                if pre_name.startswith('/'):
                    pre_name = pre_name[1:]
                return pre_name.replace('/', '-')

            data = request.data.decode('utf-8')
            item = xml_to_dataclass(data)
            if not isinstance(item, m.FunctionSetAssignments):
                _log.error("FuncitonSetAssignments was not passed via data.")
                return Response(status=400)

            if request.method == 'POST':
                if item.href:
                    _log.error(f"POST method with existing object {item.href}")
                    return Response(400)

                item = adpt.FunctionSetAssignmentsAdapter.add(item)
                response_status = 201

            elif request.method == 'PUT':
                if not item.href:
                    _log.error(f"PUT method without an existing object.")
                    return Response(400)

                index = int(item.href.rsplit(hrefs.SEP)[-1])
                adpt.FunctionSetAssignmentsAdapter.put(index, item)
                response_status = 200

            return Response(dataclass_to_xml(item), status=response_status)

        # Get all end devices
        start = int(request.args.get('s', 0))
        after = int(request.args.get('a', 0))
        limit = int(request.args.get('l', 1))

        allofem = adpt.FunctionSetAssignmentsAdapter.fetch_all(m.FunctionSetAssignmentsList(),
                                                               start=start,
                                                               after=after,
                                                               limit=limit)
        return Response(dataclass_to_xml(allofem), status=200)

    def _admin_enddevices(self) -> Response:
        """Returns EndDevice or EndDeviceList depending on the request method.
        
        If request method is GET, return an EndDeviceList object.
        
        If request method is POST, save the EndDevice object and return it with new href for the end device.
        
        If request method is PUT, update the EndDevice object and return it.
        
        If request method is DELETE, remove the EndDevice object and return None.
        """
        if request.method in ('POST', 'PUT'):

            def normalize_certificate_name(pre_name: str) -> str:
                if pre_name.startswith('/'):
                    pre_name = pre_name[1:]
                return pre_name.replace('/', '-')

            data = request.data.decode('utf-8')
            item = xml_to_dataclass(data)
            if not isinstance(item, m.EndDevice):
                _log.error("EndDevice was not passed via data.")
                return Response(status=400)

            if request.method == 'POST':
                if item.href:
                    _log.error(f"POST method with existing object {item.href}")
                    return Response(400)

                item = adpt.EndDeviceAdapter.add(item)
                cert_filename = normalize_certificate_name(item.href)
                tls_repo: TLSRepository = g.TLS_REPOSITORY
                cert, key = tls_repo.get_file_pair(cert_filename)
                Path(cert).unlink(missing_ok=True)
                Path(key).unlink(missing_ok=True)
                tls_repo.create_cert(cert_filename)

                item.lFDI = get_lfdi_from_cert(cert)
                item.sFDI = get_sfdi_from_lfdi(item.lFDI)
                index = int(item.href.rsplit(hrefs.SEP)[-1])
                adpt.EndDeviceAdapter.put(index, item)
                create_device_capability(index)
                response_status = 201

            elif request.method == 'PUT':
                if not item.href:
                    _log.error(f"PUT method without an existing object.")
                    return Response(400)

                index = int(item.href.rsplit(hrefs.SEP)[-1])
                adpt.EndDeviceAdapter.put(index, item)
                response_status = 200

            return Response(dataclass_to_xml(item), status=response_status)

        # Get all end devices
        start = int(request.args.get('s', 0))
        after = int(request.args.get('a', 0))
        limit = int(request.args.get('l', 1))

        allofem = adpt.EndDeviceAdapter.fetch_all(m.EndDeviceList(),
                                                  start=start,
                                                  after=after,
                                                  limit=limit)
        return Response(dataclass_to_xml(allofem), status=200)

    def _admin_controls(self) -> Response:
        """Returns DERControl or DERControlList depending on the request method.
        
        If request method is GET, return a DERControlList object.
        
        If request method is POST, save the DERControl object and return it with new href for the control.
        
        If request method is PUT, update the DERControl object and return it.
        
        If request method is DELETE, remove the DERControl object and return None.
        
        """

        if request.method in ('POST', 'PUT'):
            data = request.data.decode('utf-8')
            control = xml_to_dataclass(data)
            if not isinstance(control, m.DERControl):
                _log.error("DERControl was not passed via data.")
                return Response(status=400)

            if request.method == 'POST':
                if control.href:
                    _log.error(f"POST method with existing object {control.href}")
                    return Response(400)

                control = adpt.DERControlAdapter.add(control)
                response_status = 201

            elif request.method == 'PUT':
                if not control.href:
                    _log.error(f"PUT method without an existing object.")
                    return Response(400)

                index = int(control.href.rsplit(hrefs.SEP)[-1])
                adpt.DERControlAdapter.put(index, control)
                response_status = 200

            return Response(dataclass_to_xml(control), status=response_status)

        start = int(request.args.get('s', 0))
        after = int(request.args.get('a', 0))
        limit = int(request.args.get('l', 1))

        return Response(dataclass_to_xml(
            adpt.DERControlAdapter.fetch_all(m.DERControlList(),
                                             start=start,
                                             after=after,
                                             limit=limit)),
                        status=200)

    def _admin_curves(self) -> Response:
        """Returns DERCurve or DERCurve List depending upon request method.
        
        If request method is GET, return a DERCurveList object.
        
        If request method is POST or PUT, return a DERCurve object.
        
        If request method is DELETE return None.        
        """
        if request.method in ('POST', 'PUT'):
            data = request.data.decode('utf-8')
            curve = xml_to_dataclass(data)
            if not isinstance(curve, m.DERCurve):
                _log.error("DERCurve was not passed via data.")
                return Response(status=400)

            if request.method == 'POST':
                if curve.href:
                    _log.error(f"POST method with existing object {curve.href}")
                    return Response(400)

                curve = adpt.DERControlAdapter.add(curve)
                response_status = 201

            elif request.method == 'PUT':
                if not curve.href:
                    _log.error(f"PUT method without an existing object.")
                    return Response(400)

                index = int(curve.href.rsplit(hrefs.SEP)[-1])
                adpt.DERControlAdapter.put(index, curve)
                response_status = 200

            return Response(dataclass_to_xml(curve), status=response_status)

        start = int(request.args.get('s', 0))
        after = int(request.args.get('a', 0))
        limit = int(request.args.get('l', 1))

        return Response(dataclass_to_xml(
            adpt.DERCurveAdapter.fetch_all(m.DERCurveList(), start=start, after=after,
                                           limit=limit)),
                        status=200)

    def _admin_edev(self) -> Response:
        return Response(dataclass_to_xml(adpt.EndDeviceAdapter.fetch_all(m.EndDeviceList())))

    def _admin_edev_ders(self, edevid: int, derid: int = None) -> Response:
        ed = EndDeviceAdapter.fetch(edevid)
        deradpter: Adapter[m.DER] = EndDeviceAdapter.fetch_child(ed, hrefs.DER)
        if derid:
            retval: m.DER = deradpter.fetch(derid)
            if request.path.endswith('current_derp'):
                derp_href = hrefs.DERProgramHrefOld.parse(retval.CurrentDERProgramLink)
                retval = DERProgramAdapter.fetch(derp_href.index)
        else:
            retval = deradpter.fetch_all(m.DERList())

        return Response(dataclass_to_xml(retval))

    def _admin_edev_fsa(self, edevid: int, fsaid: int = -1) -> Response:
        if not edevid > -1:
            raise ValueError("Invalid end device id passed")

        ed = EndDeviceAdapter.fetch(edevid)

        if fsaid > -1:
            obj = FSAAdapter.fetch(fsaid)
        else:
            obj = FSAAdapter.fetch_all(m.FunctionSetAssignmentsList())

        # if edevid > -1 and fsaid > -1:
        #     obj = EndDeviceAdapter.fetch_all(m.FunctionSetAssignmentsList())
        #     #obj = EndDeviceAdapter.fetch_fsa(edev_index=edevid, fsa_index=fsaid)
        # elif edevid > -1:
        #     ed = EndDeviceAdapter.fetch(edevid)
        #     obj = EndDeviceAdapter.fetch_child(ed, hrefs.FSA)
        #     #obj = EndDeviceAdapter.fetch_fsa_list(edev_index=edevid)
        # else:
        #     return Response("Invalid edevid specified", status=400)

        return Response(dataclass_to_xml(obj))

    def _admin_edev_fsa_derp(self, edevid: int, fsaid: int) -> Response:
        derps = EndDeviceAdapter.fetch_derp_list(edev_index=edevid, fsa_index=fsaid)
        return Response(dataclass_to_xml(derps))

    def _admin_der_program_lists(self) -> Response:
        return Response(dataclass_to_xml(DERProgramAdapter.fetch_list()))

    # def _admin_ders(self, edev_index: int) -> Response:
    #     return Response(dataclass_to_xml(DERAdapter.fetch_list(edev_index=edev_index)))

    # def _der_settings(self, edev_index: int, der_index: int):
    #     return Response(dataclass_to_xml(DERAdapter.fetch_settings_at(edev_index=edev_index, der_index=der_index)))

    def _admin_der_update_current_derp(self, edev_index: int, der_index: int):
        if request.method == 'PUT':
            data: m.DERProgram = xml_to_dataclass(request.data.decode('utf-8'))
            if not isinstance(data, m.DERProgram):
                return Response(status=400)

            if data.mRID:
                program = DERProgramAdapter.fetch_by_mrid(data.mRID)
                response_status = 200
                if not program:
                    program = DERProgramAdapter.create(data).data
                    response_status = 201
            else:
                program = DERProgramAdapter.create(data).data
                response_status = 201
            print(EndDeviceAdapter.fetch_child_names())
            der = DERAdapter.fetch_at(edev_index, der_index)
            der.CurrentDERProgramLink = m.CurrentDERProgramLink(program.href)
            return Response(dataclass_to_xml(program), status=response_status)
        else:
            der = DERAdapter.fetch_at(edev_index, der_index)
            if der.CurrentDERProgramLink:
                parsed = hrefs.der_program_parse(der.CurrentDERProgramLink.href)
                program = DERProgramAdapter.fetch_at(parsed.index)
            else:
                program = m.DERProgram()

            return Response(dataclass_to_xml(program))

    def _admin_derp(self, index: int = -1) -> Response:

        # Get all end devices
        start = int(request.args.get('s', 0))
        after = int(request.args.get('a', 0))
        limit = int(request.args.get('l', 1))

        if request.method == 'GET':
            if index < 0:
                retval = adpt.ListAdapter.get_resource_list(hrefs.DEFAULT_DERP_ROOT,
                                                            start=start,
                                                            after=after,
                                                            limit=limit)
                #return Response(dataclass_to_xml(DERProgramAdapter.fetch_all(m.DERProgramList())))
            elif request.method == 'GET':
                retval = adpt.ListAdapter.get(hrefs.DEFAULT_DERP_ROOT, index)

            return Response(dataclass_to_xml(retval))

        if request.method == 'POST':
            raise NotImplemented("POST not implemented")
            # xml = request.data.decode('utf-8')
            # data = xml_to_dataclass(request.data.decode('utf-8'))

            # if not isinstance(data, m.DERProgram):
            #     raise BadRequest("Invalid DERProgram")
            # response = DERProgramAdapter.create(data)

            return Response(headers={'Location': response.href}, status=response.statusint)

        return Response(f"I am {index}, {request.method}")

    def _admin_derp_derca(self, derp_index: int) -> Response:
        ctrl_list = DERProgramAdapter.fetch_der_active_control_list(derp_index)
        return Response(dataclass_to_xml(ctrl_list))

    def _admin_derp_derc(self, derp_index: int) -> Response:
        derp = adpt.ListAdapter.get(hrefs.DEFAULT_DERP_ROOT, derp_index)
        # derp = adpt.DERProgramAdapter.fetch(derp_index)

        if request.method in ("PUT", "POST"):

            data = request.get_data(as_text=True)
            # Retrieve data from xml
            data = xml_to_dataclass(data)

            # If we are retrieving the default instance.
            if isinstance(data, m.DefaultDERControl):
                data.href = derp.DefaultDERControlLink.href

                status_code = 201
                if get_href(derp.DefaultDERControlLink.href):
                    status_code = 204
                add_href(data.href, data)

            elif isinstance(data, m.DERControl):
                status_code = 201

                der_cntl_list: m.DERControlList = adpt.ListAdapter.get_resource_list(
                    derp.DERControlListLink.href)
                assert isinstance(der_cntl_list, m.DERControlList)

                found_at = None
                for indx, cntl in enumerate(der_cntl_list.DERControl):
                    if cntl.mRID == data.mRID:
                        found_at = indx

                if not found_at:
                    data.href = derp.DERControlListLink.href + hrefs.SEP + str(
                        len(der_cntl_list.DERControl))
                    adpt.ListAdapter.append(derp.DERControlListLink.href, data)
                    #der_cntl_list.DERControl.append(data)
                else:
                    status_code = 204
                    data.href = der_cntl_list.DERControl[found_at].href
                    der_cntl_list.DERControl[found_at] = data

                der_cntl_list.all = len(der_cntl_list.DERControl)
                #add_href(derp.DERControlListLink.href, der_cntl_list)
                adpt.TimeAdapter.add_event(data)

            return Response(headers={'Location': data.href}, status=status_code)

    def _admin(self) -> Response:
        arg_path = request.args.get('path')
        device = request.args.get('device')

        if arg_path == '/enddevices':
            return Response(dataclass_to_xml(EndDeviceAdapter.fetch_list()))

        if arg_path.startswith('/edev'):
            edev_path = hrefs.edev_parse(arg_path)
            if edev_path.der_index == hrefs.NO_INDEX:
                return Response(
                    dataclass_to_xml(DERAdapter.fetch_list(edev_index=edev_path.edev_index)))
            elif edev_path.der_sub is None:
                return Response(
                    dataclass_to_xml(
                        DERAdapter.fetch_at(edev_index=edev_path.edev_index,
                                            der_index=edev_path.der_index)))
            elif edev_path.der_sub == hrefs.DERSubType.CurrentProgram.value:
                return Response(
                    dataclass_to_xml(
                        DERAdapter.fetch_current_program_at(edev_index=edev_path.edev_index,
                                                            der_index=edev_path.der_index)))

        elif arg_path.startswith("/fsa"):

            fsa_path = hrefs.fsa_parse(arg_path)

            if fsa_path.fsa_index == hrefs.NO_INDEX:
                return Response(dataclass_to_xml(FSAAdapter.fetch_list()))
            elif fsa_path.fsa_index != hrefs.NO_INDEX and fsa_path.fsa_sub is None:
                return Response(dataclass_to_xml(FSAAdapter.fetch_at(fsa_path.fsa_index)))
            else:
                return Response(dataclass_to_xml(FSAAdapter.fetch_program_list(
                    fsa_path.fsa_index)))

    def _lfdi_lists(self) -> Response:
        items = []

        for k, v in self.end_devices.__all_end_devices__.items():
            items.append({"key": k, "lfdi": int(v.end_device.lFDI)})

        return Response(json.dumps(items))

    # def _admin_edev_fsa(self, edevid: int, fsaid: int = -1) -> Response:
    #     #edev = self.end_devices.get(edevid)
    #     return Response(json.dumps(json.dumps(self.end_devices.get_fsa_list(edevid=edevid))))

    def _program_lists(self) -> str:
        return render_template("admin/program-lists.html",
                               program_lists=self.server_config.program_lists)

    def _admin_enddevice_list(self) -> str:
        return render_template("admin/end-device-list.html",
                               end_device_list=self.end_devices.get_end_devices())



from __future__ import annotations
import logging
from dataclasses import dataclass
from typing import Dict, Callable, Optional

import werkzeug
from flask import request, Response

from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.models import DeviceCategoryType
import ieee_2030_5.server.server_endpoints as eps

from ieee_2030_5.types_ import SEP_XML
from ieee_2030_5.utils import dataclass_to_xml

_log = logging.getLogger(__name__)


class ServerOperation:

    def __init__(self):
        if 'ieee_2030_5_peercert' not in request.environ:
            raise werkzeug.exceptions.Forbidden()
        self._headers = {'Content-Type': SEP_XML}

    def head(self, **kwargs):
        raise werkzeug.exceptions.MethodNotAllowed()

    def get(self, **kwargs):
        raise werkzeug.exceptions.MethodNotAllowed()

    def post(self, **kwargs):
        raise werkzeug.exceptions.MethodNotAllowed()

    def delete(self, **kwargs):
        raise werkzeug.exceptions.MethodNotAllowed()

    def put(self, **kwargs):
        raise werkzeug.exceptions.MethodNotAllowed()

    def execute(self, **kwargs):
        methods = {
            'GET': self.get,
            'POST': self.post,
            'DELETE': self.delete,
            'PUT': self.put
        }

        fn = methods.get(request.environ['REQUEST_METHOD'])
        if not fn:
            raise werkzeug.exceptions.MethodNotAllowed()

        return fn(**kwargs)


class RequestOp(ServerOperation):
    def __init__(self, server_endpoints: eps.ServerEndpoints):
        super().__init__()
        self._tls_repository = server_endpoints.tls_repo
        self._server_endpoints = server_endpoints

    @property
    def tls_repo(self) -> TLSRepository:
        return self._tls_repository

    @property
    def server_config(self) -> ServerConfiguration:
        return self._server_endpoints.config

    @property
    def lfdi(self):
        return request.environ["ieee_2030_5_lfdi"] # self._tls_repository.lfdi(request.environ['ieee_2030_5_subject'])

    @property
    def device_id(self):
        return request.environ.get("ieee_2030_5_subject")

    def get_path(self, required_prefix: Optional[str] = None) -> str:
        """
        Retrieve the context web request environment PATH_INFO with optional required_prefix
        argument.  If that argument is specified then it will be validated against PATH_INFO.  The
        function will raise a ValueError if the PATH_INFO does not start with required_prefix.

        Args:
            required_prefix:

        Returns:
            The path specified in request.environ['PATH_INFO'
        """

        pth = request.environ['PATH_INFO']

        if required_prefix and not pth.startswith(required_prefix):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")

        return pth

    def build_response_from_dataclass(self, obj: dataclass) -> Response:
        return Response(dataclass_to_xml(obj), headers=self._headers)



import logging
import os
import urllib.parse
from enum import Enum

import requests

import ieee_2030_5.models as m
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass

_log = logging.getLogger(__name__)

backend_session = requests.Session()
ADMIN_URL = ""

class SaveRequestMethod(Enum):
    POST = "POST"
    PUT = "PUT"
    
def setup_backend_session():
    global backend_session, ADMIN_URL
    
    if not ADMIN_URL:    
        backend_session.cert = (os.getenv("2030_5_CLIENT_CERT"), os.getenv("2030_5_CLIENT_KEY"))
        backend_session.verify = os.getenv("2030_5_CA_CERT")
        ADMIN_URL = f"https://{os.getenv('2030_5_HOST')}:{os.getenv('2030_5_PORT')}/admin" 

def list_endpoint(endpoint: str, start: int = 0, after: int = 0, limit: int = 0) -> str:
    setup_backend_session()
    base_url = ADMIN_URL
    while endpoint.startswith('/'):
        endpoint = endpoint[1:]
    endpoint = urllib.parse.quote(endpoint)
    endpoint += f"?s={start}&a={after}&l={limit}"
    return f"{base_url}/{endpoint}"
        

def list_parameters(start: int = 0, after: int = 0, limit: int = 0):
    return dict(s=start, a=after, l=limit)

def endpoint(endpoint: str) -> str:
    setup_backend_session()
    base_url = ADMIN_URL
    while endpoint.startswith('/'):
        endpoint = endpoint[1:]
    endpoint = urllib.parse.quote(endpoint)
    return f"{base_url}/{endpoint}"

def get_der_list() -> m.DERList:
    href = endpoint('der')
    list_params = list_parameters()
    return xml_to_dataclass(backend_session.get(href, params=list_params).text)

def send_der(item: m.DER) -> m.DER:
    item_xml = dataclass_to_xml(item)
        
    if item.href:
        _log.debug("PUTTING data")
        response = backend_session.put(endpoint('der'), data=item_xml)
    else:
        _log.debug("POSTING data")
        response = backend_session.post(endpoint('der'), data=item_xml)
    
    _log.debug(response.text)
    
    return xml_to_dataclass(response.text)

def get_enddevice_list() -> m.EndDeviceList:
    href = endpoint('enddevice')
    list_params = list_parameters()
    return xml_to_dataclass(backend_session.get(href, params=list_params).text)

def send_enddevice(item: m.EndDevice) -> m.EndDevice:
    item_xml = dataclass_to_xml(item)
        
    if item.href:
        _log.debug("PUTTING data")
        response = backend_session.put(endpoint('enddevices'), data=item_xml)
    else:
        _log.debug("POSTING data")
        response = backend_session.post(endpoint('enddevices'), data=item_xml)
    
    _log.debug(response.text)
    
    return xml_to_dataclass(response.text)

def get_fsa_list() -> m.FunctionSetAssignmentsList:
    href = endpoint('fsa')
    list_params = list_parameters()
    return xml_to_dataclass(backend_session.get(href, params=list_params).text)

def send_fsa(fsa: m.FunctionSetAssignments) -> m.FunctionSetAssignments:
    item_xml = dataclass_to_xml(fsa)
        
    if fsa.href:
        _log.debug("PUTTING data")
        response = backend_session.put(endpoint('fsa'), data=item_xml)
    else:
        _log.debug("POSTING data")
        response = backend_session.post(endpoint('fsa'), data=item_xml)
    
    _log.debug(response.text)
    
    return xml_to_dataclass(response.text)

def get_program_list() -> m.DERProgramList:
    href = endpoint('programs')
    list_params = list_parameters()
    return xml_to_dataclass(backend_session.get(href, params=list_params).text)

def send_program(program: m.DERProgram) -> m.DERProgram:
    slug = "programs"
    item = program
    item_xml = dataclass_to_xml(program)
    
        
    if item.href:
        _log.debug(f"PUTTING {item.__class__.__name__} data: {item_xml}")
        response = backend_session.put(endpoint(slug), data=item_xml)
    else:
        _log.debug(f"POSTING {item.__class__.__name__} data: {item_xml}")
        response = backend_session.post(endpoint(slug), data=item_xml)
    
    _log.debug(response.text)
    
    return xml_to_dataclass(response.text)


def get_control_list() -> m.DERControlList:
    href = endpoint('controls')
    list_params = list_parameters()
    return xml_to_dataclass(backend_session.get(href, params=list_params).text)


def send_control(control: m.DERControl) -> m.DERControl:
    slug = "controls"
    item = control
    item_xml = dataclass_to_xml(control)
        
    if item.href:
        _log.debug(f"PUTTING {item.__class__.__name__} data: {item_xml}")
        response = backend_session.put(endpoint(slug), data=item_xml)
    else:
        _log.debug(f"POSTING {item.__class__.__name__} data: {item_xml}")
        response = backend_session.post(endpoint(slug), data=item_xml)
    
    _log.debug(response.text)
    
    return xml_to_dataclass(response.text)

def get_curve_list() -> m.DERCurveList:
    href = endpoint('curves')
    list_params = list_parameters()
    return xml_to_dataclass(backend_session.get(href, params=list_params).text)

def send_curve(curve: m.DERCurve) -> m.DERCurve:
    slug = "curves"
    item = curve
    item_xml = dataclass_to_xml(curve)
        
    if curve.href:
        _log.debug("PUTTING data")
        response = backend_session.put(endpoint(slug), data=item_xml)
    else:
        _log.debug("POSTING data")
        response = backend_session.post(endpoint(slug), data=item_xml)
    
    _log.debug(response.text)
    
    return xml_to_dataclass(response.text)

def get_enddevice_list() -> m.EndDeviceList:
    return xml_to_dataclass(backend_session.get(endpoint('enddevices')).text)

def get_cert_list():
    certs = backend_session.get(endpoint('certs')).json()
    return certs


from __future__ import annotations

import io
import logging
import os
import sys

from dotenv import load_dotenv
from fastapi import FastAPI
from nicegui import ui

from .pages import Pages, show_global_header

logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
_log = logging.getLogger(__name__)

app = FastAPI()

__all__ = [
    "run_gui"
]

def initialize_app():
    
    show_global_header(Pages.HOME)
    
    ui.markdown("""

# Welcome to the IEEE 2030.5 Admin GUI

This is a work in progress""")
    
    # device_list = get_end_devices()
    # certs = get_certs()

    # print(certs)
    # print(device_list)
    # row_data = []
    # for ed in device_list.EndDevice:
    #     row = dict(name=f'{ed.lFDI}',
    #             cert=f'<a href="/admin/cert/{ed.lFDI}">cert</a>')
    #     row_data.append(row)
        

    # ui.aggrid({
    #     'columnDefs': [
    #         {'headerName': '', 'field': 'cert'},
    #         {'headerName': 'Name', 'field': 'name'},
    #     ],
    #     'rowData': row_data,
    # }, html_columns=[0])
def run_gui():
    out_file = open(sys.stdout.fileno(), 'wb', 0)
    sys.stdout = io.TextIOWrapper(out_file, write_through=True)
    from pathlib import Path
    
    PREFIX = "2030_5"
    
    # Allow custmization of the 2030_5_ENV_PATH
    env_path = Path(os.environ.get(f"{PREFIX}_ENV_PATH", 
                                   str(Path(__file__).parent / ".env"))) 
    
    
    load_dotenv(dotenv_path=env_path.as_posix())

    required_env = [
        f"{PREFIX}_HOST",
        f"{PREFIX}_PORT",
        f"{PREFIX}_CLIENT_KEY",
        f"{PREFIX}_CLIENT_CERT",
        f"{PREFIX}_CA_CERT"
    ]
    
    for x in required_env:
        if os.getenv(x) is None or os.getenv(x).strip() == "":
            sys.stderr.write(f'Missing {x} in .env file\n')
            sys.exit(1)
            
    logging.basicConfig(filename="gui-server.log", level=logging.DEBUG)
    
    initialize_app()
    
    ui.run(show=False)
    # from pathlib import Path
    
    # env_path = Path(__file__).parent / ".env"
    
    # load_dotenv(dotenv_path=env_path.as_posix())

    # PREFIX = "2030_5"
    # required_env = [
    #     f"{PREFIX}_HOST",
    #     f"{PREFIX}_PORT",
    #     f"{PREFIX}_CLIENT_KEY",
    #     f"{PREFIX}_CLIENT_CERT",
    #     f"{PREFIX}_CA_CERT"
    # ]
    
    # for x in required_env:
    #     if os.getenv(x) is None or os.getenv(x).strip() == "":
    #         sys.stderr.write(f'Missing {x} in .env file\n')
    #         sys.exit(1)
            
    # logging.basicConfig(level=logging.DEBUG)
    
    # initialize_app()
    
    
    # ui.run(show=False)
    
    
            
    

# ui.html("""
#     <table>
#         <tr>
#             <td>1</td>
#             <td>2</td>
#         </tr>
#     </table>
# """)


# columns = [
#     {'name': 'name', 'label': 'Name', 'field': 'name', 'required': True},
#     {'name': 'age', 'label': 'Age', 'field': 'age', 'sortable': True},
# ]
# rows = [
#     {'id': 0, 'name': 'Alice', 'age': 18},
#     {'id': 1, 'name': 'Bob', 'age': 21},
#     {'id': 2, 'name': 'Lionel', 'age': 19},
#     {'id': 3, 'name': 'Michael', 'age': 32},
#     {'id': 4, 'name': 'Julie', 'age': 12},
#     {'id': 5, 'name': 'Livia', 'age': 25},
#     {'id': 6, 'name': 'Carol'},
# ]

# with ui.table(title='My Team', columns=columns, rows=rows, selection='multiple', pagination=10).classes('w-96') as table:
#     with table.add_slot('top-right'):
#         with ui.input(placeholder='Search').props('type=search').bind_value(table, 'filter').add_slot('append'):
#             ui.icon('search')
#     with table.add_slot('body'):
#         with table.row() as row:
#             with table.cell() as cell:
#                 pprint(row.props)
#                 ui.link('Download', f'https://nicegui.io?{row.props.id}')
#             with table.cell():
#                 ui.link('Download', 'https://nicegui.io')
#             with table.cell():
#                 ui.link('Download', 'https://nicegui.io')
#     with table.add_slot('bottom-row'):
#         with table.row():
#             with table.cell():
#                 ui.button(on_click=lambda: (
#                     table.add_rows({'id': time.time(), 'name': new_name.value, 'age': new_age.value}),
#                     new_name.set_value(None),
#                     new_age.set_value(None),
#                 ), icon='add').props('flat fab-mini')
#             with table.cell():
#                 new_name = ui.input('Name')
#             with table.cell():
#                 new_age = ui.number('Age')

# ui.label().bind_text_from(table, 'selected', lambda val: f'Current selection: {val}')
# ui.button('Remove', on_click=lambda: table.remove_rows(*table.selected)) \
#     .bind_visibility_from(table, 'selected', backward=lambda val: bool(val))

    #ui.run(show=False)


"""
This module handles MirrorUsagePoint and UsagePoint constructs for a server.
"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional

from flask import Response, request
from werkzeug.exceptions import BadRequest

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.data.indexer import get_href
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.server.uuid_handler import UUIDHandler
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass


class Error(Exception):
    pass


@dataclass
class ResponseStatus:
    location: str
    status: str


class UsagePointRequest(RequestOp):

    def get(self) -> Response:

        start = int(request.args.get("s", 0))
        limit = int(request.args.get("l", 1))
        after = int(request.args.get("a", 0))
        parsed = hrefs.ParsedUsagePointHref(request.path)

        handled = False
        sort_by = []
        reversed = True

        if parsed.has_reading_list():
            sort_by = "timePeriod.start"
            if handled := parsed.reading_index is not None:
                obj = adpt.ListAdapter.get(parsed.last_list(), parsed.reading_index)

        elif parsed.has_reading_set_list():
            sort_by = "timePeriod.start"
            if handled := parsed.reading_set_index is not None:
                obj = adpt.ListAdapter.get(parsed.last_list(), parsed.reading_set_index)

        elif parsed.has_meter_reading_list():
            if handled := parsed.has_reading_type():
                obj = get_href(request.path)
            elif handled := parsed.meter_reading_index is not None:
                obj = adpt.ListAdapter.get(parsed.last_list(), parsed.meter_reading_index)
        else:
            obj = adpt.ListAdapter.get_resource_list(request.path,
                                                     start=start,
                                                     limit=limit,
                                                     after=after,
                                                     reverse=reversed)

        if not handled:
            obj = adpt.ListAdapter.get_resource_list(request.path,
                                                     start=start,
                                                     limit=limit,
                                                     after=after,
                                                     sort_by=sort_by,
                                                     reverse=reversed)
        # # /upt
        # if not parsed.has_usage_point_index():
        #     obj = adpt.UsagePointAdapter.fetch_all(m.UsagePointList(request.path),
        #                                            start=start,
        #                                            limit=limit,
        #                                            after=after)
        # elif parsed.has_meter_reading_list() and not parsed.meter_reading_index:
        #     obj = adpt.MirrorUsagePointAdapter.fetch_all(m.MeterReadingList(request.path),
        #                                                  start=start,
        #                                                  limit=limit,
        #                                                  after=after)

        # else:
        #     obj = adpt.UsagePointAdapter.fetch(parsed.usage_point_index)

        # if parsed.has_extra():
        #     obj = get_href(request.path)

        return self.build_response_from_dataclass(obj)


class MirrorUsagePointRequest(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        pth_info = request.path

        if not pth_info.startswith(hrefs.DEFAULT_MUP_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")

        mup_href = hrefs.ParsedUsagePointHref(request.path)

        if not mup_href.has_usage_point_index():
            # /mup
            mup: m.MirrorUsagePointList = adpt.ListAdapter.get_resource_list(request.path)
            # Because our resource_list doesn't include other properties than the list we set
            # them here before returning.
            mup.pollRate = self.server_config.mirror_usage_point_post_rate
        else:
            # /mup_0
            mup = adpt.ListAdapter.get(hrefs.DEFAULT_MUP_ROOT, mup_href.usage_point_index)

        return self.build_response_from_dataclass(mup)

    def post(self) -> Response:
        xml = request.data.decode('utf-8')
        data = xml_to_dataclass(request.data.decode('utf-8'))
        data_type = type(data)
        if data_type not in (m.MirrorUsagePoint, m.MirrorReadingSet, m.MirrorMeterReading):
            raise BadRequest()

        pth_info = request.path
        pths = pth_info.split(hrefs.SEP)
        if len(pths) == 1 and data_type is not m.MirrorUsagePoint:
            # Check to make sure not a new mrid
            raise BadRequest("Must post MirrorUsagePoint to top level only")

        # Creating a new mup
        if data_type == m.MirrorUsagePoint:
            if data.postRate is None:
                data.postRate = self.server_config.mirror_usage_point_post_rate
            result = adpt.create_mirror_usage_point(mup=data)
            #result = adpt.MirrorUsagePointAdapter.create(mup=data)
        else:
            result = adpt.create_mirror_meter_reading(mup_href=request.path, mmr_input=data)

        if result.success:
            status = '204' if result.was_update == True else '201'
        else:
            status = '405'

        if status.startswith('20'):
            if result.location:
                return Response(headers={'Location': result.location}, status=status)
            return Response(headers={'Location': result.an_object.href}, status=status)
        else:
            return Response(result.an_object, status=status)



from typing import Optional

from flask import Response, request
from werkzeug.exceptions import NotFound

import ieee_2030_5.adapters as adpt
from ieee_2030_5.data.indexer import add_href, get_href
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.utils import xml_to_dataclass

import logging

_log = logging.getLogger(__name__)


class DERRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def put(self) -> Response:
        """Allows putting of 2030.5 DER data to the server.
        """
        if not request.path.startswith(hrefs.DEFAULT_DER_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")

        parser = hrefs.HrefParser(request.path)

        clstype = {
            hrefs.DER_SETTINGS: m.DERSettings,
            hrefs.DER_STATUS: m.DERStatus,
            hrefs.DER_CAPABILITY: m.DERCapability,
            hrefs.DER_AVAILABILITY: m.DERAvailability,
            hrefs.DER_PROGRAM: m.DERProgram,
        }

        data = request.get_data(as_text=True)
        data = xml_to_dataclass(data, clstype[parser.at(2)])

        _log.debug(f"DER PUT {request.path} {data}")

        #orig = get_href(request.path)
        # if orig is None:
        #     raise NotFound(f"{request.path}")
        # data.href = orig.href
        add_href(request.path, data)
        return self.build_response_from_dataclass(data)

    def get(self) -> Response:

        if not request.path.startswith(hrefs.DEFAULT_DER_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")

        value = get_href(request.path)

        if value is None:

            parser = hrefs.HrefParser(request.path)

            subpaths = {
                hrefs.DER_SETTINGS: m.DERSettings(href=request.path),
                hrefs.DER_STATUS: m.DERStatus(href=request.path),
                hrefs.DER_CAPABILITY: m.DERCapability(href=request.path),
                hrefs.DER_AVAILABILITY: m.DERAvailability(href=request.path),
                hrefs.DER_PROGRAM: m.DERProgram(href=request.path),
            }

            if parser.has_index():
                index = parser.at(1)
                subpath = parser.at(2)
                value = subpaths[subpath]

        # pth_split = request.path.split(hrefs.SEP)

        # if len(pth_split) == 1:
        #     # TODO Add arguments
        #     value = adpt.DERAdapter.fetch_list()
        # else:
        #     value = adpt.DERAdapter.fetch_at(int(pth_split[1]))

        return self.build_response_from_dataclass(value)


class DERProgramRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:

        start = int(request.args.get('s', 0))
        after = int(request.args.get('a', 0))
        limit = int(request.args.get('l', 1))

        parsed = hrefs.HrefParser(request.path)

        if not parsed.has_index():
            retval = adpt.ListAdapter.get_resource_list(hrefs.DEFAULT_DERP_ROOT, start, after,
                                                        limit)
        elif parsed.count() == 2:
            retval = adpt.ListAdapter.get(hrefs.DEFAULT_DERP_ROOT, parsed.at(1))
        elif parsed.count() == 4:
            # Retrive the list of controls from storage
            dercl = get_href(parsed.join(3))
            assert isinstance(dercl, m.DERControlList)
            # The index that we want to get the control from.
            retval = dercl.DERControl[parsed.at(3)]
        elif parsed.at(2) == hrefs.DERC:
            retval = adpt.ListAdapter.get_resource_list(request.path, start, after, limit)
        # elif parsed.at(2) == hrefs.DDERC:
        #     retval = adpt.DERControlAdapter.fetch_at(parsed.at(3))
        else:
            retval = get_href(request.path)

        if not retval:
            raise NotFound(f"{request.path}")

        return self.build_response_from_dataclass(retval)



from ast import Dict
from pathlib import Path
import readline
import shlex
from typing import List

import yaml

import ieee_2030_5.models as m

data = {}

class CommandError(Exception):
    pass

def menu():
    
    print(f"""
Options are as follows:

{"help":<40} Show this message

Sections:
  {"controls":<38} Enter control editing mode
  {"devices":<38} Enter device editing mode
  {"programs":<38} Enter program editing mode

Commands:
  {"restore":<38} Restore from the cached config in case of a crash
  {"load config <filename>":<38} Load a config from yaml
  {"store config <filename>":<38} Store config <filename> to yaml
  {"quit (q)":<38} Exit Program
""")

    
def check_loaded(data):
    if not data:
        raise CommandError(f"Load the config before {cmd}")


def load_config(filename) -> Dict:
    pth = Path(filename)
    if not pth.exists():
        raise CommandError(f"Invalid filename ('{filename}')")
    data = yaml.safe_load(pth.read_text())
    print(f"Loaded {filename}")
    return data


def store_config(filename, data):
    pth = Path(filename)
    with pth.open('w') as fs:
        yaml.dump(data, fs) 
    print(f"Stored {filename}")


def print_list(heading, items: List):
    print(heading)
    print("-" * 30)
    for item in items:
        print(item)
            

def print_devices(devices):
    print_list("Devices", devices)
        

def print_programs(programs):
    print_list("Programs", programs) 
    

def print_controls(controls):
    print_list("Controls", controls)
          

def handle_controls():
    global data
    controls = data.get('controls', [])
    if not controls:
        data['controls'] = controls
    default_prompt = "controls>"
    prompt = default_prompt
    current_item = -1
    while True:
        
        cmd, *args = shlex.split(input(prompt))
        try:
            if cmd == "q":                
                if prompt == default_prompt:
                    break
                else:
                    prompt = default_prompt
            elif cmd == "new":
                try:
                    description, *other = args
                except ValueError:
                    raise CommandError("Must specify a description")
                if other:
                    raise CommandError("Only a description passed to new")
                controls.append(dict(description=description))
                prompt = f"{default_prompt}{len(controls)}>{description}>"
                current_item = len(controls) - 1
            elif cmd == "set":
                args
                key, value = args
                controls[current_item][key] = value
            elif cmd in ("ls", "show"):
                print_controls(controls)
            else:
                raise CommandError(f"Invalid command specified {cmd}")
                
        except CommandError as ex:
            print(ex)
        finally:
            store_working()

def handle_device_settings(device_index: int):
    global data
    settings = data["devices"][device_index].get("settings", {})
    if not settings:
        data["devices"][device_index]["settings"] = settings
        
    default_prompt = f"devices>{data['devices'][device_index]['id']}>settings>"
    prompt = default_prompt
    
    while True:
        
        cmd, *args = shlex.split(input(prompt))
        try:
            if cmd in ("q", "quit"):
                if prompt == default_prompt:
                    break
                else:
                    prompt = default_prompt
            elif cmd == "set":
                try:
                    key, value = args
                    settings[key] = value
                except ValueError:
                    raise CommandError("Invalid arguments passed to set.")
        except CommandError as ex:
            print(ex.args[0])

def handle_end_devices():
    global data
    end_devices = data.get('devices', [])
    if not end_devices:
        data['devices'] = end_devices
    default_prompt = "devices>"
    prompt = default_prompt
    current_item = -1
    while True:
        
        cmd, *args = shlex.split(input(prompt))
        try:
            if cmd in ("q", "quit"):
                if prompt == default_prompt:
                    break
                else:
                    prompt = default_prompt
            elif cmd == "new":
                try:
                    id, *other = args
                except ValueError:
                    raise CommandError("Must specify an id")
                if other:
                    raise CommandError("Only a id passed to new")
                
                found_at = [index for index, dev in enumerate(end_devices) if dev['id'] == id]
                if found_at:
                    current_item = [found_at]
                else:
                    end_devices.append(dict(id=id))
                    current_item = len(end_devices) -1
                prompt = f"{default_prompt}{len(end_devices)}>{id}>"
                
            elif cmd == "set":
                key, value = args
                end_devices[current_item][key] = value
            elif cmd in ("ls", "show"):
                print_controls(end_devices)
            elif cmd == "settings":
                try:
                    current_item = int(args[0])
                    if current_item > len(end_devices) - 1 or current_item < 0:
                        raise CommandError("Invalid end device index")
                    
                    handle_device_settings(current_item)
                except (IndexError, ValueError):
                    raise CommandError("Missing index to end device")
                
            else:
                raise CommandError(f"Invalid command specified {cmd}")
                
        except CommandError as ex:
            print(ex)
        finally:
            store_working()
        

def handle_programs():
    global data
    devices = data.get("devices", [])
    if not devices:
        data["devices"] = devices
    controls = data.get('controls', [])
    if not controls:
        data['controls'] = controls
    programs = data.get('programs', [])
    if not programs:
        data['programs'] = programs
    default_prompt = "programs>"
    prompt = default_prompt
    current_item = -1
    while True:
        
        cmd, *args = shlex.split(input(prompt))
        try:
            if cmd == "q":
                if prompt == default_prompt:
                    break
                else:
                    prompt = default_prompt
            elif cmd == "new":
                try:
                    description, *other = args
                except ValueError:
                    raise CommandError("Must specify a description")
                if other:
                    raise CommandError("Only a description passed to new")
                programs.append(dict(description=description))
                prompt = f"programs>{len(programs)}>{description}>"
                current_item = len(programs) - 1
            elif cmd == "set":
                key, value = args
                programs[current_item][key] = value
            elif cmd == "control":
                try:
                    op, index, *other = args
                    
                    index = int(index)
                    if index >= len(controls):
                        raise CommandError(f"Invalid control index specified for default control {index}")
                    if op == "default":
                        programs[current_item]["default_control"] = index
                    elif op == "add":
                        current_controls = programs[current_item].get("controls", [])
                        if not current_controls:
                            programs[current_item]["controls"] = current_controls
                        if index in current_controls:
                            raise CommandError("Index already present in controls")
                        programs[current_item]["controls"].append(index)
                        
                except ValueError:
                    raise CommandError("Specify index of control to be default") 
            elif cmd in ("ls", "show"):
                print_programs(programs)
            else:
                raise CommandError(f"Invalid command specified {cmd}")
                
        except CommandError as ex:
            print(ex)
        finally:
            store_working()
    

def store_working():
    pth = Path("/tmp/.working_data")
    with pth.open('w') as fs:
        yaml.dump(data, fs)


def load_working():
    pth = Path("/tmp/.working_data")
    if pth.exists():
        with pth.open('r') as fs:
            working = yaml.safe_load(fs)
            data.update(working)

def print_all():
    global_items = [{"id": k, "value": data[k]} for k in sorted(data) if data not in ('devices', 'controls', 'programs')]
    print_list("Global", global_items)
    print_controls(data.get('controls', []))
    print_devices(data.get('devices', []))
    print_programs(data.get('programs', []))
    
    
def _main():
    global data
    
    file_type = ""
    
    while True:
        cmd, *args = shlex.split(input('> '))
        try:
            if cmd == "q" or cmd == "quit":
                break
            elif cmd == "help":
                menu()
            elif cmd == "load":
                file_type, file = args 
                if file_type == "config":
                    data = load_config(file)
                else:
                    raise CommandError(f"Invalid file_type {file_type}")
            elif cmd == "store":
                check_loaded(data)
                new_file_type, file = args 
                if new_file_type == "config":
                    store_config(file, data)
                else:
                    raise CommandError(f"Invalid file_type {new_file_type}")
            elif cmd == "devices":
                handle_end_devices()
            elif cmd == "programs":
                handle_programs()
            elif cmd == "controls":
                handle_controls()
            elif cmd == "show":
                print_all()
            elif cmd == "set":
                try:
                    key, value = args
                except ValueError:
                    raise CommandError("Not enough arguments to set.  set key value is the format.")
                    
                available_keys = ['proxy_hostname', 'server_hostname', "server_mode", "tls_repository"]
                if key not in available_keys:
                    raise CommandError(f"{key} was not in the available list of properties that can be set.")
                data[key] = value
            elif cmd == "restore":
                load_working()
            else:
                raise CommandError(f"Invalid command {cmd}")
            
        except CommandError as ex:
            print(ex.args[0])                 
        finally:
            store_working()
if __name__ == '__main__':
    _main()
        




from dataclasses import dataclass
from enum import Enum
from typing import Callable

from nicegui import ui


@dataclass
class PageContext:
    name: str
    title: str
    uri: str
    module: Callable = None
    
    

class Pages(Enum):
    HOME = PageContext('home', 'Home', '/')
    CERTS = PageContext('certs', 'Certificates', '/certs')
    CURVES = PageContext('curves', 'Curves', '/curves')
    CONTROLS = PageContext('controls', 'DER Controls', '/controls')
    DEFAULT_CONTROLS = PageContext('default_controls', 'Default DER Controls', '/default_controls')
    ENDDEVICES = PageContext('enddevices', 'End Devices', '/enddevices')
    PROGRAMS = PageContext('programs', 'DER Programs', '/programs')
    FSA = PageContext('fsa', 'Function Set Assignments', '/fsa')
    DER = PageContext('der', 'DER', '/der')
    # LOGIN = 'login'
    # LOGOUT = 'logout'
    # SETTINGS = 'settings'
    # USERS = 'users'


def show_global_header(page: PageContext):
    
    with ui.header(elevated=True).style('background-color: #3874c8'): #.classes('justify-between'):
        for index, pg in enumerate(Pages):
            link = ui.link(pg.value.title, pg.value.uri).style('color: white')
            if pg.value == page:
                link.style('font-weight: bold')

from .certs import show_certs
from .controls import show_controls
from .curves import show_curves
from .der import show_der
from .enddevices import show_enddevices
from .fsa import show_fsa
from .programs import show_programs

Pages.CERTS.value.module = show_certs
Pages.CONTROLS.value.module = show_controls
Pages.CURVES.value.module = show_curves
Pages.ENDDEVICES.value.module = show_enddevices
Pages.PROGRAMS.value.module = show_programs
Pages.FSA.value.module = show_fsa
Pages.DER.value.module = show_der

# def load_pages():
#     import glob
#     from os.path import basename, dirname, isfile, join
#     modules = glob.glob(join(dirname(__file__), "*.py"))
#     __all__ = [ basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]
    


from __future__ import annotations

#import io
import logging
import os
import sys

from dotenv import load_dotenv
#from fastapi import FastAPI
from nicegui import ui


logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
_log = logging.getLogger(__name__)

# @ui.page("/")
# def run_page():
#     ui.markdown("""Ignoring this""")

# ui.run()

# app = FastAPI()

# __all__ = [
#     "run_gui"
# ]
def setup_environment():
    """Creates the os.environ for the application.
    
    This is done using the .env file in same directory as this file or from
    the command line parameters or from the environmental variables already existing.
    """
    from pathlib import Path

    PREFIX = "2030_5"

    # Allow custmization of the 2030_5_ENV_PATH
    env_path = Path(os.environ.get(f"{PREFIX}_ENV_PATH", 
                                    str(Path(__file__).parent / ".env"))) 


    load_dotenv(dotenv_path=env_path.as_posix())

    required_env = [
        f"{PREFIX}_HOST",
        f"{PREFIX}_PORT",
        f"{PREFIX}_CLIENT_KEY",
        f"{PREFIX}_CLIENT_CERT",
        f"{PREFIX}_CA_CERT"
    ]

    for x in required_env:
        if os.getenv(x) is None or os.getenv(x).strip() == "":
            sys.stderr.write(f'Missing {x} in .env file\n')
            sys.exit(1)
    
    paths = [
        f"{PREFIX}_CLIENT_KEY",
        f"{PREFIX}_CLIENT_CERT",
        f"{PREFIX}_CA_CERT"
    ]
    
    for x in paths:
        os.environ[x] = Path(os.getenv(x)).expanduser().as_posix()

def initialize_app():
    
    show_global_header(Pages.HOME)
    
    ui.markdown("""

# Welcome to the IEEE 2030.5 Admin GUI

This is a work in progress""")
    
if __name__ in {"__main__", "__mp_main__"}:
    # out_file = open(sys.stdout.fileno(), 'wb', 0)
    # sys.stdout = io.TextIOWrapper(out_file, write_through=True)
    
    # Must be done before importing the rest of the application.
    setup_environment()

    from ieee_2030_5_gui.pages import Pages, show_global_header            
    
    logging.basicConfig(level=logging.DEBUG)

    initialize_app()
    
    # excludes = ['data_store/**',
    #             'ieee_2030_5/adapters/**',
    #             'ieee_2030_5/client', 
    #             'ieee_2030_5/data]
    includes = ['ieee_2030_5_gui/**',]

    ui.run(show=False, uvicorn_reload_includes=",".join(includes))
    
    # from pathlib import Path
    
    # env_path = Path(__file__).parent / ".env"
    
    # load_dotenv(dotenv_path=env_path.as_posix())

    # PREFIX = "2030_5"
    # required_env = [
    #     f"{PREFIX}_HOST",
    #     f"{PREFIX}_PORT",
    #     f"{PREFIX}_CLIENT_KEY",
    #     f"{PREFIX}_CLIENT_CERT",
    #     f"{PREFIX}_CA_CERT"
    # ]
    
    # for x in required_env:
    #     if os.getenv(x) is None or os.getenv(x).strip() == "":
    #         sys.stderr.write(f'Missing {x} in .env file\n')
    #         sys.exit(1)
            
    # logging.basicConfig(level=logging.DEBUG)
    
    # initialize_app()
    
    
    # ui.run(show=False)
    
    
            


# ui.html("""
#     <table>
#         <tr>
#             <td>1</td>
#             <td>2</td>
#         </tr>
#     </table>
# """)


# columns = [
#     {'name': 'name', 'label': 'Name', 'field': 'name', 'required': True},
#     {'name': 'age', 'label': 'Age', 'field': 'age', 'sortable': True},
# ]
# rows = [
#     {'id': 0, 'name': 'Alice', 'age': 18},
#     {'id': 1, 'name': 'Bob', 'age': 21},
#     {'id': 2, 'name': 'Lionel', 'age': 19},
#     {'id': 3, 'name': 'Michael', 'age': 32},
#     {'id': 4, 'name': 'Julie', 'age': 12},
#     {'id': 5, 'name': 'Livia', 'age': 25},
#     {'id': 6, 'name': 'Carol'},
# ]

# with ui.table(title='My Team', columns=columns, rows=rows, selection='multiple', pagination=10).classes('w-96') as table:
#     with table.add_slot('top-right'):
#         with ui.input(placeholder='Search').props('type=search').bind_value(table, 'filter').add_slot('append'):
#             ui.icon('search')
#     with table.add_slot('body'):
#         with table.row() as row:
#             with table.cell() as cell:
#                 pprint(row.props)
#                 ui.link('Download', f'https://nicegui.io?{row.props.id}')
#             with table.cell():
#                 ui.link('Download', 'https://nicegui.io')
#             with table.cell():
#                 ui.link('Download', 'https://nicegui.io')
#     with table.add_slot('bottom-row'):
#         with table.row():
#             with table.cell():
#                 ui.button(on_click=lambda: (
#                     table.add_rows({'id': time.time(), 'name': new_name.value, 'age': new_age.value}),
#                     new_name.set_value(None),
#                     new_age.set_value(None),
#                 ), icon='add').props('flat fab-mini')
#             with table.cell():
#                 new_name = ui.input('Name')
#             with table.cell():
#                 new_age = ui.number('Age')

# ui.label().bind_text_from(table, 'selected', lambda val: f'Current selection: {val}')
# ui.button('Remove', on_click=lambda: table.remove_rows(*table.selected)) \
#     .bind_visibility_from(table, 'selected', backward=lambda val: bool(val))

    #ui.run(show=False)


import logging
import os
import sys
from pathlib import Path

import uvicorn
from fastapi import FastAPI
from fastapi.responses import RedirectResponse

from ieee_2030_5_gui.init import init

logging.basicConfig(level=logging.DEBUG)

api = FastAPI()

@api.get("/")
def read_root():
    return  RedirectResponse(url='/certs')

init(api)

if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    
    parser.add_argument("config", help="Path to config server configuration file.")
    
    opts = parser.parse_args()
    
    # Because we are going to run uvicorn in a subprocess, we need to pass the config file to it.
    # We do that here through creating a .env file that uvicorn will read.
    with open('.env', 'w') as f:
        f.write(f"CONFIG_FILE={opts.config}")
    
    uvicorn.run("main:api", host="0.0.0.0", port=8000, reload=True, env_file='.env')


import random
import uuid
from datetime import datetime, timedelta

from nicegui import app, ui
from router import global_router as router
from session import backend_session, endpoint

import ieee_2030_5.models as m
from ieee_2030_5.utils import dataclass_to_xml, uuid_2030_5, xml_to_dataclass


def datachange(data, prop, x):
    if hasattr(x, "value"):
        value = x.value
    else:
        value = x
    setattr(data, prop, value)
    return getattr(data, prop)

def add_control_event(program_index: int):
    der_control = m.DERControl()
    
    
    def post_der_control():
        resp = backend_session.post(endpoint(f'/derp/{program_index}/derc'), dataclass_to_xml(der_control))
        if resp.ok:
            ui.open("/derp")
        
        
        
    with ui.column():
        with ui.row():
            der_control.mRID = uuid_2030_5()
            ui.input('mRID', value=der_control.mRID, on_change=lambda e: datachange(der_control, "mRID", e))
        with ui.row():
            currentdate = datetime.utcnow()
            fmtstring = '%Y-%m-%d %H:%M:%S'
            ui.label(f"Current time UTC: {currentdate.strftime(fmtstring)}")
        with ui.row():
            ui.input("Event Start Date Time", value=f"{(datetime.utcnow() + timedelta(seconds=90)).strftime(fmtstring)}", on_change=lambda e: datachange(der_control, "interval.start", e))
        
        with ui.row():            
            duration = ui.input("duration", on_change=lambda e: datachange(der_control, "interval.duration", int(e.value)))
        ui.button("Save", on_click=lambda: post_der_control())

@router.add('/derp')
def programs_list():
    resp = backend_session.get(endpoint('/derp'))
    derps: m.DERProgramList = xml_to_dataclass(resp.text)
        
    for index, derp in enumerate(derps.DERProgram):
        router.add_route(f"/derp/{index}/derc/add", lambda: add_control_event(index))
        with ui.row():
            with ui.column():
                ui.label(f"Description: {derp.description}")
                ui.label(f"mRID: {derp.mRID}")                
                ui.label(f"primacy: {derp.primacy}")
                
                if derp.ActiveDERControlListLink:
                    ui.label(derp.ActiveDERControlListLink.href)
                if derp.DefaultDERControlLink:
                    ui.label(derp.DefaultDERControlLink.href)
                if derp.DERControlListLink:
                    ui.label(derp.DERControlListLink.href)
                
                ui.button('Control Event', on_click=lambda: router.open(f"/derp/{index}/derc/add"))

    ui.button('Create New Program', on_click=lambda: router.open(f'/derp/add'))
    
@app.get('/derp/{max}')
def generate_random_number(max: int):
    with ui.column():
        ui.label("foo")
    #return {'min': 0, 'max': max, 'value': random.randint(0, max)}
  

@router.add('/derp/add')
def program_add():
    data = m.DERProgram()
    
    
    def create_derp():
        resp = backend_session.post(endpoint("derp"), data=dataclass_to_xml(data))
        if resp.ok:
            ui.open("/derp")
        
    with ui.column():
        ui.input("mRID", on_change=lambda x: datachange(data, "mRID", x))
        ui.input("description", on_change=lambda x: datachange(data, "description", x))
        ui.input("primacy", on_change=lambda x: datachange(data, "primacy", x))
        ui.button("Save", on_click=lambda: create_derp())

#ui.run()




from dataclasses import dataclass, field
import threading
from typing import Callable, Dict, List, Union
from ieee_2030_5.data.indexer import add_href
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
import ieee_2030_5.adapters as adpt
from ieee_2030_5.adapters import Adapter, NotFoundError, ResourceListAdapter
from ieee_2030_5.config import ReturnValue

from datetime import datetime
import time

from blinker import Signal
import logging

_log = logging.getLogger(__name__)

DERCurveAdapter = Adapter[m.DERCurve](hrefs.curve_href(), generic_type=m.DERCurve)
DERControlAdapter = Adapter[m.DERControl]("/derc", generic_type=m.DERControl)
DERCurveAdapter = Adapter[m.DERCurve](hrefs.curve_href(), generic_type=m.DERCurve)
DERProgramAdapter = Adapter[m.DERProgram](hrefs.der_program_href(), generic_type=m.DERProgram)
FunctionSetAssignmentsAdapter = Adapter[m.FunctionSetAssignments](
    url_prefix="/fsa", generic_type=m.FunctionSetAssignments)

EndDeviceAdapter = Adapter[m.EndDevice](hrefs.get_enddevice_href(), generic_type=m.EndDevice)
DeviceCapabilityAdapter = Adapter[m.DeviceCapability]("/dcap", generic_type=m.DeviceCapability)
# Generally the href will only be in the context of an end device.
RegistrationAdapter = Adapter[m.Registration](url_prefix="/reg", generic_type=m.Registration)
DERAdapter = Adapter[m.DER](url_prefix="/der", generic_type=m.DER)
MirrorUsagePointAdapter = Adapter[m.MirrorUsagePoint](url_prefix="/mup",
                                                      generic_type=m.MirrorUsagePoint)
MirrorMeterReadingAdapter = Adapter[m.MirrorMeterReading](url_prefix="/mr",
                                                          generic_type=m.MirrorMeterReading)
MirrorReadingSetAdapter = Adapter[m.MirrorReadingSet](url_prefix="/rs",
                                                      generic_type=m.MirrorReadingSet)
UsagePointAdapter = Adapter[m.UsagePoint](url_prefix="/upt", generic_type=m.UsagePoint)
ListAdapter = ResourceListAdapter()


def create_mirror_meter_reading(
        mup_href: str, mmr_input: Union[m.MirrorMeterReading,
                                        m.MirrorMeterReadingList]) -> ReturnValue:

    if isinstance(mmr_input, m.MirrorMeterReadingList):
        raise NotImplemented()

    mup: m.MirrorUsagePoint = adpt.ListAdapter.get(hrefs.DEFAULT_MUP_ROOT,
                                                   mup_href.split(hrefs.SEP)[-1])
    assert isinstance(mup, m.MirrorUsagePoint)
    assert len(mup.MirrorMeterReading) == 1

    was_updated = False
    location = None
    if isinstance(mmr_input, m.MirrorMeterReading):
        # Attempt to find an existing mup with the same mrid.  If found then we need to replace
        # the data with the new data etc.  If not found then we add the mmr to the list of mmrs.
        mmr_list_href = hrefs.SEP.join([mup_href, "mr"])
        mr_list_href = mmr_list_href.replace("mup", "upt")
        try:
            mmr = adpt.ListAdapter.get_by_mrid(mmr_list_href, mmr_input.mRID)
            mmr_index = adpt.ListAdapter.get_list(mmr_list_href).index(mmr)
            was_updated = True
        except NotFoundError:
            mmr_index = adpt.ListAdapter.list_size(mmr_list_href)
            mmr_input.href = hrefs.SEP.join([mmr_list_href, str(mmr_index)])

            for mup_mr in mup.MirrorMeterReading:
                mmr_input.ReadingType = mup_mr.ReadingType
                mmr_input.description = mup_mr.description
                mmr_input.lastUpdateTime = mup_mr.lastUpdateTime
            ListAdapter.append(mmr_list_href, mmr_input)
            mmr = mmr_input

        try:
            mr = adpt.ListAdapter.get(mr_list_href, mmr_index)
        except NotFoundError:
            # This shouldn't happen if it does then there is something wrong with our code.
            if was_updated:
                raise ValueError("Unable to find meter reading for updated mirror meter reading")
            rt_href = hrefs.SEP.join([mr_list_href, str(mmr_index), "rt"])
            mmr.ReadingType.href = rt_href
            mr = m.MeterReading(href=hrefs.SEP.join([mr_list_href, str(mmr_index)]),
                                mRID=mmr.mRID,
                                description=mmr.description,
                                ReadingTypeLink=m.ReadingTypeLink(rt_href))
            add_href(rt_href, mmr.ReadingType)
            ListAdapter.append(mr_list_href, mr)
        location = mr_list_href

        # Current instantanious values.
        if mmr_input.Reading is not None:
            mmr.Reading.href = hrefs.SEP.join([mmr.href, "r"])
            mmr.Reading.href = mmr.Reading.href.replace("mup", "upt")
            add_href(mmr.Reading.href, mmr.Reading)
            mr.ReadingLink = m.ReadingLink(mmr.Reading.href)

        # Mirror reading sets
        if mmr_input.MirrorReadingSet:

            mrs_list_href = hrefs.SEP.join([mmr.href, "rs"])
            rs_list_href = hrefs.SEP.join([mmr.href, "rs"]).replace("mup", "upt")
            mr.ReadingSetListLink = m.ReadingSetListLink(href=rs_list_href)
            ListAdapter.initialize_uri(mr.ReadingSetListLink.href, m.ReadingSet)

            for mrs in mmr_input.MirrorReadingSet:
                found_rs = False
                try:
                    mrs_item = ListAdapter.get_by_mrid(mrs_list_href, mrs.mRID)
                    mrs_item_index = ListAdapter.get_list(mrs_list_href).index(mrs_item)
                    found_rs = True
                except NotFoundError:
                    mrs_item = mrs
                    mrs_item_index = ListAdapter.list_size(mrs_list_href)
                    mrs_item.href = hrefs.SEP.join([mrs_list_href, str(mrs_item_index)])
                    ListAdapter.append(mrs_list_href, mrs_item)

                if found_rs:
                    rs_item = ListAdapter.get(rs_list_href, mrs_item_index)
                    rs_item.description = mrs_item.description
                    rs_item.timePeriod = mrs_item.timePeriod
                    rs_item.version = mrs_item.version
                else:
                    rs_item = m.ReadingSet(href=hrefs.SEP.join([rs_list_href,
                                                                str(mrs_item_index)]),
                                           description=mrs_item.description,
                                           timePeriod=mrs_item.timePeriod,
                                           version=mrs_item.version)
                    ListAdapter.append(rs_list_href, rs_item)

                reading_list_href = hrefs.SEP.join([rs_item.href, "r"])
                rs_item.ReadingListLink = m.ReadingListLink(href=reading_list_href)
                for reading_index, reading in enumerate(mrs_item.Reading):
                    reading.href = hrefs.SEP.join([reading_list_href, str(reading_index)])
                    ListAdapter.append(reading_list_href, reading)

    ListAdapter.store()

    return ReturnValue(True, mmr, was_updated, location)


def create_mirror_usage_point(mup: m.MirrorUsagePoint) -> ReturnValue:
    """Creates a MirrorUsagePoint and associated UsagePoint and adds them to their adapters.
    """

    # Attempt to find an existing mup with the same mrid.  If found then we need to replace
    # it with the new data etc.
    found_with_mrid = None
    if adpt.ListAdapter.list_size(hrefs.DEFAULT_MUP_ROOT) > 0:
        try:
            found_with_mrid = adpt.ListAdapter.get_by_mrid(hrefs.DEFAULT_MUP_ROOT, mup.mRID)
        except NotFoundError:
            ...

    update = False
    if not found_with_mrid:
        # Creating a new allocation of resources for the mup. And then copy the data from the
        # mup resources into the usage point resources allocating new data as needed.
        upt_href = hrefs.UsagePointHref()

        # Both the usage point and mirror usage point will have the same point index.
        point_index = adpt.ListAdapter.list_size(hrefs.DEFAULT_MUP_ROOT)
        mup.href = hrefs.SEP.join([hrefs.DEFAULT_MUP_ROOT, str(point_index)])
        # Add the mirror usage point to the list of mirror usage points.
        adpt.ListAdapter.append(hrefs.DEFAULT_MUP_ROOT, mup)

        # Create a usage point with the same index as the mirror usage point.
        upt = m.UsagePoint(href=hrefs.SEP.join([hrefs.DEFAULT_UPT_ROOT,
                                                str(point_index)]),
                           description=mup.description,
                           deviceLFDI=mup.deviceLFDI,
                           serviceCategoryKind=mup.serviceCategoryKind,
                           mRID=mup.mRID,
                           roleFlags=mup.roleFlags,
                           status=mup.status)
        upt.MeterReadingListLink = m.MeterReadingListLink(href=hrefs.SEP.join([upt.href, "mr"]))
        adpt.ListAdapter.append(hrefs.DEFAULT_UPT_ROOT, upt)

        # Initialize the url for the mirror meter reading list.
        mmr_list_href = hrefs.SEP.join([mup.href, "mr"])
        adpt.ListAdapter.initialize_uri(mmr_list_href, m.MirrorMeterReading)

        # Initialize the url for the meter reading list.
        mr_list_href = hrefs.SEP.join([upt.href, "mr"])
        adpt.ListAdapter.initialize_uri(mr_list_href, m.MeterReading)

        for index_for_readings, mirror_meter_reading in enumerate(mup.MirrorMeterReading):
            # Validate the the reading has a reading type associated with it.
            if not mirror_meter_reading.ReadingType:
                return ReturnValue(
                    False,
                    f"Invalid Reading Type for Mirror Meter Reading {mirror_meter_reading.mRID}")

            # Update the mirror meter reading href and then add it to the list of mirror meter readings.
            mirror_meter_reading.href = hrefs.SEP.join([mmr_list_href, str(index_for_readings)])
            adpt.ListAdapter.append(mmr_list_href, mirror_meter_reading)

            # Create a meter reading for the usage point.
            meter_reading = m.MeterReading(href=hrefs.SEP.join(
                [mr_list_href, str(index_for_readings)]),
                                           mRID=mirror_meter_reading.mRID,
                                           description=mirror_meter_reading.description)
            adpt.ListAdapter.append(mr_list_href, meter_reading)

            # Store the type of reading using the add_href method.
            rt_href = hrefs.SEP.join([upt.href, "rt", str(index_for_readings)])
            meter_reading.ReadingTypeLink = m.ReadingTypeLink(rt_href)
            add_href(rt_href, mirror_meter_reading.ReadingType)

    else:
        # TODO Update all properties with new items from mup
        mup.href = found_with_mrid.href
        found_with_mrid.description = mup.description
        found_with_mrid.deviceLFDI = mup.deviceLFDI
        found_with_mrid.serviceCategoryKind = mup.serviceCategoryKind
        found_with_mrid.mRID = mup.mRID
        found_with_mrid.roleFlags = mup.roleFlags
        found_with_mrid.status = mup.status
        update = True
        adpt.ListAdapter.store()

    return ReturnValue(True, mup, update, mup.href)


@dataclass
class TimerSpec:
    trigger_after_seconds: int
    fn: Callable
    args: List = field(default_factory=list)
    kwargs: Dict = field(default_factory=dict)
    enabled: bool = True
    trigger_count: int = 0
    last_trigger_time: int = int(time.mktime(datetime.utcnow().timetuple()))

    def disable(self):
        self.enabled = False

    def enable(self):
        self.enabled = True

    def reset_count(self):
        self.trigger_count = 0

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TimerSpec):
            raise NotImplementedError(
                f"Comparison between {self.__class__.__name__} and {type(other)} not implemented")
        return self.fn is other.fn

    def trigger(self, current_time: int):
        if self.last_trigger_time + self.trigger_after_seconds < current_time:
            if self.args and self.kwargs:
                self.fn(args=self.args, kwargs=self.kwargs)
            elif self.args:
                self.fn(args=self.args)
            else:
                self.fn()
            self.trigger_count += 1
            self.last_trigger_time = current_time


class _TimeAdapter(threading.Thread):
    tick = Signal("tick")
    event_started = Signal("event_started")
    event_ended = Signal("event_endend")
    event_scheduled = Signal("event_scheduled")
    events: Dict[str, m.Event] = {}
    current_tick: int = 0
    looping: bool = False

    @staticmethod
    def user_readable(timestamp: int) -> str:
        dt = datetime.fromtimestamp(timestamp)
        return dt.isoformat()    # .strftime("%m/%d/%Y, %H:%M:%S")

    @staticmethod
    def from_iso(iso_fmt_date: str) -> int:
        dt = datetime.strptime(iso_fmt_date, "%Y-%m-%dT%H:%M:%S")
        return int(time.mktime(dt.timetuple()))

    @staticmethod
    def add_event(evnt: m.Event):
        time_now = _TimeAdapter.current_tick
        if evnt.EventStatus is None:
            evnt.EventStatus = m.EventStatus()
        if evnt.href not in _TimeAdapter.events:
            while _TimeAdapter.looping:
                time.sleep(0.1)
            _TimeAdapter.events[evnt.href] = evnt

    def run(self) -> None:

        while True:
            _TimeAdapter.current_tick = int(time.mktime(datetime.utcnow().timetuple()))
            _TimeAdapter.tick.send(self.current_tick)
            _TimeAdapter.looping = True
            time_now = _TimeAdapter.current_tick
            for href, evnt in _TimeAdapter.events.items():
                if time_now < evnt.interval.start and evnt.EventStatus.currentStatus is None:
                    evnt.EventStatus.dateTime = time_now
                    evnt.EventStatus.currentStatus = 0
                    _log.debug(f"{'='*20}Event Scheduled {evnt.href}")
                    _TimeAdapter.event_scheduled.send(evnt)
                elif evnt.interval.start < time_now and time_now < evnt.interval.start + evnt.interval.duration:
                    if evnt.EventStatus.currentStatus != 1:
                        evnt.EventStatus.currentStatus = 1
                        evnt.EventStatus.dateTime = time_now
                        _log.debug(f"{'='*20}Event Started {evnt.href}")
                        _TimeAdapter.event_started.send(evnt)
                elif time_now > evnt.interval.start + evnt.interval.duration and evnt.EventStatus.currentStatus == 1:
                    evnt.EventStatus.currentStatus = 5
                    evnt.EventStatus.dateTime = time_now
                    _log.debug(f"{'='*20}Event Complete {evnt.href}")
                    _TimeAdapter.event_ended.send(evnt)

            _TimeAdapter.looping = False
            time.sleep(1)


TimeAdapter = _TimeAdapter()
TimeAdapter.daemon = True
TimeAdapter.start()


# from ieee_2030_5.server.exceptions import AlreadyExistsError
# from ieee_2030_5.server.server_constructs import (
#     Group,
#     GroupLevel,
#     get_groups,
#     get_group,
#     get_der_program_list,
#     create_group
# )
# from ieee_2030_5.server.uuid_handler import UUIDHandler
#
# from ieee_2030_5.server.base_request import (
#     ServerOperation,
#     RequestOp
# )
#
# __all__ = [
#     "Group",
#     "GroupLevel",
#     "get_groups",
#     "get_group",
#     "create_group",
#     "get_der_program_list",
#     "ServerOperation",
#     "UUIDHandler",
#     "AlreadyExistsError",
#     "RequestOp"
# ]
from flask import Request as FlaskRequest


class Request(FlaskRequest):
    pass



"""

"""
import logging
import os
from contextvars import ContextVar
from pathlib import Path
from typing import Optional

import yaml

from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration

_log = logging.getLogger(__name__)

cfg_file = Path(os.environ.get("CONFIG_FILE")).expanduser().resolve(strict=True)

if not cfg_file.exists():
    raise FileNotFoundError(f"Config file {cfg_file} does not exist")

_log.debug(f"Loading config from {cfg_file}")
cfg_dict = yaml.safe_load(cfg_file.read_text())

_server_config = ServerConfiguration(**cfg_dict)
_tls_repo = TLSRepository(_server_config.tls_repository,
                            _server_config.openssl_cnf,
                            _server_config.server_hostname,
                            _server_config.proxy_hostname,
                            clear=False)

_log.debug(f"Loading context vars for tls and server config")
tls_repo_var = ContextVar("tls_repo_var", default=_tls_repo)
server_config_var = ContextVar("server_config_var", default=_server_config)

def get_tls_repo() -> TLSRepository:
    tls_repo = tls_repo_var.get()
    assert tls_repo, "TLS Repo not loaded"
    return tls_repo

def get_server_config() -> ServerConfiguration:
    server_config = server_config_var.get()
    assert server_config, "Server config not loaded"
    return server_config





from fastapi import FastAPI
from nicegui import ui

from .pages.certs import show_certs
from .pages.controls import show_controls


def init(fastapi_app: FastAPI) -> None:
    
    @fastapi_app.get("/admin/cert")
    
    # @ui.page('/show')
    # def show():
    #     ui.label('Hello, FastAPI!')

    #     # NOTE dark mode will be persistent for each user across tabs and server restarts
    #     ui.dark_mode().bind_value(app.storage.user, 'dark_mode')
    #     ui.checkbox('dark mode').bind_value(app.storage.user, 'dark_mode')

    ui.run_with(
        fastapi_app,
        storage_secret='pick your private secret here',  # NOTE setting a secret is optional but allows for persistent storage per user
    )


"""
Provides a key/value store interface for setting retrieving points from a datastore.

This implementation just uses pickledb for loading a json datastore.  After
each set the data will be written to disk.
"""
import atexit
from pathlib import Path

from simplekv.fs import FilesystemStore

db = FilesystemStore(Path("~/.ieee_2030_5_data").expanduser().resolve())


def set_point(key: str, value: bytes):
    """
    Set a point into the key/value store.  Both key and value must be hashable types.

    Example:
        set_point("_e55a4c7a-c006-4596-b658-e23bc771b5cb.angle", -156.38295096513662)
        set_point("known_mrids": ["_4da919f1-762f-4755-b674-5faccf3faec6"])
    """
    k = key.replace('/', '^^^^')
    db.put(k, value)


def get_point(key):
    """
    Retrieve a point from the key/value store.  If the key doesn't exist returns None.
    """
    k = key.replace('/', '^^^^')
    return db.get(k)


def get_hrefs():
    return db.keys()


if __name__ == '__main__':

    set_point("foo", b"bar")
    set_point("bim", b"baf")
    # print(get_point("foo"))
    get_hrefs()



from __future__ import annotations

import logging
import os
import ssl
from dataclasses import dataclass
from http.client import HTTPSConnection
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Tuple
from urllib.parse import urlparse

import OpenSSL
import yaml

from ieee_2030_5.certs import (TLSRepository, lfdi_from_fingerprint,
                               sfdi_from_lfdi)
from ieee_2030_5.config import ServerConfiguration

_log = logging.getLogger(__name__)


@dataclass
class ContextWithPaths:
    context: ssl.SSLContext
    certpath: str
    keypath: str


class RequestForwarder(BaseHTTPRequestHandler):

    def get_context_cert_pair(self) -> ContextWithPaths:
        x509_binary = self.connection.getpeercert(True)
        cert_file = None
        key_file = None
        if x509_binary:
            x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1, x509_binary)
            # subject = x509.get_subject().CN
            # serial_num = x509.get_serial_number()
            # fingerprint = x509.digest("sha256").decode("ascii")
            # _log.info(f"Connected using fingerprint: {fingerprint}")
            # lfdi = lfdi_from_fingerprint(fingerprint)
            # _log.info(f"LFDI -> {lfdi}")
            # sfdi = sfdi_from_lfdi(lfdi)
            # _log.info(f"SFDI -> {sfdi}")
            cn = x509.get_subject().CN
            cert_file, key_file = self.server.tls_repo.get_file_pair(cn)

        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        context.verify_mode = ssl.CERT_OPTIONAL

        # context.load_verify_locations(cafile="/home/gridappsd/tls/certs/ca.crt")
        assert cert_file
        ca_file = str(Path(cert_file).parent.joinpath("ca.crt"))
        context.load_verify_locations(cafile=ca_file)
        if cert_file is not None and key_file is not None and \
                Path(cert_file).exists() and Path(key_file).exists():
            context.load_cert_chain(certfile=cert_file, keyfile=key_file)
        return ContextWithPaths(context=context, certpath=cert_file, keypath=key_file)

    def __start_request__(self) -> HTTPSConnection:
        ccp = self.get_context_cert_pair()

        host, port = self.server.proxy_target
        conn = HTTPSConnection(host=host, port=port, context=ccp.context)
        conn.connect()
        return conn

    def __handle_response__(self, conn: HTTPSConnection):

        response = conn.getresponse()
        data = response.read()

        _log.debug(f"Response from server:\n{data}")
        self.wfile.write(f'HTTP/1.1 {response.status}\n'.encode('utf-8'))

        for k, v in response.headers.items():

            if k not in ('Connection', ):
                if k == 'Content-Length':
                    self.send_header(k, str(len(data)))
                else:
                    self.send_header(k, v)
        self.end_headers()

        # self.wfile.write(data)
        # self.send_response(response.status, data.decode("utf-8"))
        self.wfile.write(data)
        self.close_connection = False
        return response

    def do_GET(self):

        conn = self.__start_request__()

        conn.request(method="GET", url=self.path, headers=self.headers, encode_chunked=True)

        response = self.__handle_response__(conn)
        _log.info(f"GET {self.path} Content-Length: {self.headers.get('Content-Length')}, Response Status: {response.status}")

    def do_POST(self):
        _log.info(f"POST {self.path} Content-Length: {self.headers.get('Content-Length')}")

        conn = self.__start_request__()
        read_data = self.rfile.read(int(self.headers.get('Content-Length')))

        conn.request(method="POST",
                     url=self.path,
                     headers=self.headers,
                     body=read_data,
                     encode_chunked=True)

        response = self.__handle_response__(conn)
        _log.info(f"POST {self.path} Content-Length: {self.headers.get('Content-Length')}, Response Status: {response.status}")

    def do_DELETE(self):
        _log.info(f"DELETE {self.path} Content-Length: {self.headers.get('Content-Length')}")

        conn = self.__start_request__()
        read_data = self.rfile.read(int(self.headers.get('Content-Length')))

        conn.request(method="DELETE",
                     url=self.path,
                     headers=self.headers,
                     body=read_data,
                     encode_chunked=True)

        response = self.__handle_response__(conn)
        _log.info(f"DELETE {self.path} Content-Length: {self.headers.get('Content-Length')}, Response Status: {response.status}")

    def do_PUT(self):
        _log.info(f"PUT {self.path} Content-Length: {self.headers.get('Content-Length')}")

        conn = self.__start_request__()
        read_data = self.rfile.read(int(self.headers.get('Content-Length')))

        conn.request(method="PUT",
                     url=self.path,
                     headers=self.headers,
                     body=read_data,
                     encode_chunked=True)

        response = self.__handle_response__(conn)
        _log.info(f"PUT {self.path} Content-Length: {self.headers.get('Content-Length')}, Response Status: {response.status}")


class ProxyServer(HTTPServer):

    def __init__(self, tls_repo: TLSRepository, proxy_target: Tuple[str, int], **kwargs):
        super().__init__(**kwargs)
        self._tls_repo = tls_repo
        self._proxy_target = proxy_target

    @property
    def proxy_target(self) -> Tuple[str, int]:
        return self._proxy_target

    @property
    def tls_repo(self) -> TLSRepository:
        return self._tls_repo


def start_proxy(server_address: Tuple[str, int], tls_repo: TLSRepository,
                proxy_target: Tuple[str, int]):
    logging.getLogger().info(f"Serving {server_address} proxied to {proxy_target}")
    RequestForwarder.protocol_version = "HTTP/1.1"
    httpd = ProxyServer(server_address=server_address,
                        proxy_target=proxy_target,
                        tls_repo=tls_repo,
                        RequestHandlerClass=RequestForwarder)
    # Since version 3.10: SSLContext without protocol argument is deprecated.
    # sslctx = ssl.SSLContext()
    sslctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

    # Use optional, because we need the cert for 2030.5 but not the admin
    # interface.
    sslctx.verify_mode = ssl.CERT_OPTIONAL
    sslctx.load_verify_locations(cafile=tls_repo.ca_cert_file)

    sslctx.check_hostname = False    # If set to True, only the hostname that matches the certificate will be accepted
    sslctx.load_cert_chain(certfile=tls_repo.server_cert_file, keyfile=tls_repo.server_key_file)
    httpd.socket = sslctx.wrap_socket(httpd.socket, server_side=True)
    httpd.serve_forever()


def build_address_tuple(hostname: str) -> Tuple[str, int]:
    """Create a Tuple[str, int] from the passed hostname.

    The hostname can be formatted using https://server:port or server:port

    :param: hostname
    """
    parsed = urlparse(hostname)

    if parsed.hostname:
        hostname = (parsed.hostname, parsed.port)
    else:
        hostname = hostname.split(":")
        hostname = (hostname[0], int(hostname[1]))
    return hostname


def _main():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument(dest="config", help="Configuration file for the server.")
    parser.add_argument("--debug",
                        action="store_true",
                        default=False,
                        help="Turns debugging on for logging of the proxy.")

    opts = parser.parse_args()

    debug_level = logging.DEBUG if opts.debug else logging.DEBUG
    logging.basicConfig(level=debug_level)

    cfg_dict = yaml.safe_load(Path(opts.config).expanduser().resolve(strict=True).read_text())

    config = ServerConfiguration(**cfg_dict)

    if config.proxy_hostname is None:
        print("Invalid proxy_hostname in config file.")

    tls_repo = TLSRepository(repo_dir=config.tls_repository,
                             openssl_cnffile_template=config.openssl_cnf,
                             serverhost=config.server_hostname,
                             proxyhost=config.proxy_hostname,
                             clear=False)

    proxy_host = build_address_tuple(config.proxy_hostname)
    server_host = build_address_tuple(config.server_hostname)

    _log.debug(f"Proxy host tuple: {proxy_host}")
    _log.debug(f"Server host tuple: {server_host}")

    start_proxy(server_address=(proxy_host[0], int(proxy_host[1])),
                tls_repo=tls_repo,
                proxy_target=(server_host[0], int(server_host[1])))


if __name__ == '__main__':
    _main()



from __future__ import annotations

import logging

from blinker import Signal

# from ieee_2030_5.adapters import BaseAdapter
from ieee_2030_5.certs import TLSRepository, lfdi_from_fingerprint
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.data.indexer import add_href, get_href
from ieee_2030_5.utils.tls_wrapper import OpensslWrapper

_log = logging.getLogger(__name__)

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m


def create_device_capability(end_device_index: int) -> m.DeviceCapability:
    """Create a device capability objecct for the passed device index

    This function does not verify that there is a device at the passed index.
    """
    dcap_href = hrefs.DeviceCapabilityHref(end_device_index)
    device_capability = m.DeviceCapability()
    device_capability = dcap_href.fill_hrefs(device_capability)

    # device_capability = m.DeviceCapability(href=str(hrefs.DERCapabilitiesHref(end_device_index)))
    # device_capability.EndDeviceListLink = m.EndDeviceListLink(href=hrefs.DEFAULT_EDEV_ROOT, all=1)
    # device_capability.DERProgramListLink = m.DERProgramListLink(href=hrefs.DEFAULT_DERP_ROOT,
    #                                                             all=0)
    device_capability.MirrorUsagePointListLink = m.MirrorUsagePointListLink(
        href=hrefs.DEFAULT_MUP_ROOT, all=0)
    device_capability.TimeLink = m.TimeLink(href=hrefs.DEFAULT_TIME_ROOT)
    device_capability.UsagePointListLink = m.UsagePointListLink(href=hrefs.DEFAULT_UPT_ROOT, all=0)

    adpt.DeviceCapabilityAdapter.add(device_capability)
    return device_capability


def add_enddevice(device: m.EndDevice) -> m.EndDevice:
    """Populates links to EndDevice resources and adds it to the EndDeviceAdapter.
    
    If the link is to a single writable (by the client) resource then create the link
    and the resource with default data.  Otherwise, the link will be to a list.  It is
    expected that the list will be populated at a later point in time in the code execution.
    
    The enddevice is added to the enddevice adapter, and the following links are created and added to the enddevice:

    - `DERListLink`: A link to the DER list for the enddevice
    - `FunctionSetAssignmentsListLink`: A link to the function set assignments list for the enddevice
    - `LogEventListLink`: A link to the log event list for the enddevice
    - `RegistrationLink`: A link to the registration for the enddevice
    - `ConfigurationLink`: A link to the configuration for the enddevice
    - `DeviceInformationLink`: A link to the device information for the enddevice
    - `DeviceStatusLink`: A link to the device status for the enddevice
    - `PowerStatusLink`: A link to the power status for the enddevice

    :param device: The enddevice to add
    :type device: m.EndDevice
    :return: The enddevice object that was added to the adapter
    :rtype: m.EndDevice
    """

    # After adding to the adapter the device will have an href associated with the Adapter Type.
    device = adpt.EndDeviceAdapter.add(device)

    # Create a link object that holds references for linking other objects to the end device.
    ed_href = hrefs.EndDeviceHref(edev_href=device.href)

    ed_href.fill_hrefs(device)

    # Store objects in the href cache for retrieval.
    add_href(ed_href.configuration, m.Configuration(href=device.ConfigurationLink.href))
    add_href(ed_href.device_information,
             m.DeviceInformation(href=device.DeviceInformationLink.href))
    add_href(ed_href.device_status, m.DeviceStatus(href=device.DeviceStatusLink.href))
    add_href(ed_href.power_status, m.PowerStatus(href=device.PowerStatusLink.href))

    device.MirrorUsagePointListLink = m.MirrorUsagePointListLink(href=hrefs.DEFAULT_MUP_ROOT,
                                                                 all=0)
    device.UsagePointListLink = m.UsagePointListLink(href=hrefs.DEFAULT_UPT_ROOT, all=0)
    adpt.ListAdapter.initialize_uri(hrefs.DEFAULT_MUP_ROOT, m.MirrorUsagePoint)
    adpt.ListAdapter.initialize_uri(hrefs.DEFAULT_UPT_ROOT, m.UsagePoint)

    return device


def update_active_der_event_started(event: m.Event):
    """Event triggered when a DERControl event starts
    
    Find the control and copy it to the ActiveDERControlList

    :param event: The control event
    :type event: m.Event
    """

    assert type(event) == m.DERControl

    href_parser = hrefs.HrefEventParser(event.href)

    program = adpt.ListAdapter.get(hrefs.DEFAULT_DERP_ROOT, href_parser.program_index)

    control_list: m.DERControlList = adpt.ListAdapter.get_resource_list(
        program.DERControlListLink.href)
    control = next(filter(lambda x: x.mRID == event.mRID, control_list.DERControl))
    control.EventStatus = event.EventStatus
    assert control.EventStatus.currentStatus == 1
    adpt.ListAdapter.append(program.ActiveDERControlListLink.href, control)

    add_href(control.href, control)
    add_href(event.href, event)
    add_href(control_list.href, control_list)

    activel: m.DERControlList = get_href(program.ActiveDERControlListLink.href)

    try:
        # TODO: if found deal with supersceded eventing.
        next(filter(lambda x: x.mRID == event.mRID, activel.DERControl))
    except StopIteration:
        activel.DERControl.append(event)
        add_href(program.ActiveDERControlListLink.href, activel)


def update_active_der_event_ended(event: m.Event):
    """Event triggered when a DERControl event ends
    
    Search over the ActiveDERControlListLink for the event that has been triggered
    and remove it from the list.

    :param event: The control event
    :type event: m.Event
    """
    assert type(event) == m.DERControl

    href_parser = hrefs.HrefEventParser(event.href)

    program = adpt.ListAdapter.get(hrefs.DEFAULT_DERP_ROOT, href_parser.program_index)

    control_list: m.DERControlList = adpt.ListAdapter.get_resource_list(
        program.DERControlListLink.href)
    control = next(filter(lambda x: x.mRID == event.mRID, control_list.DERControl))
    control.EventStatus = event.EventStatus
    add_href(control.href, control)
    add_href(event.href, event)
    add_href(control_list.href, control_list)

    activel: m.DERControlList = get_href(program.ActiveDERControlListLink.href)

    remove = []
    for index, ctl in enumerate(activel.DERControl):
        if ctl.mRID == event.mRID:
            if event.EventStatus.currentStatus != 1:
                remove.insert(0, index)

    for x in remove:
        activel.DERControl.pop(x)

    add_href(program.ActiveDERControlListLink.href, activel)


adpt.TimeAdapter.event_started.connect(update_active_der_event_started)
adpt.TimeAdapter.event_ended.connect(update_active_der_event_ended)


def initialize_2030_5(config: ServerConfiguration, tlsrepo: TLSRepository):
    """Initialize the 2030.5 server.  
    
    This method initializes the adapters from the configuration objects into
    the persistence adapters.
    
    The adapters are:
    
     - EndDeviceAdapter
     - DERAdapter
     - DERCurveAdapter
     - DERProgramAdapter
     - FunctionSetAssignmentsAdapter
     
    The EndDevices in the EndDeviceAdapter will link to Lists of other types.  Those
    Lists will be stored in the ListAdapter object under the List's href (see below /edev_0_der).
    As an example the following, note the DER href is not /edev_0_der_0, but /der_12 instead.
    
    <EndDevice href="/edev_0">
      <DERList href="/edev_0_der" all="1">
    </EndDevice>
    <DERList href="/edev_0_der" all="1" result="1">
      <DER href="/der_12">
        ...
      </DER>
    </DERList
        
    
    """
    _log.debug("Initializing 2030.5")
    _log.debug("Adding server level urls to cache")

    end_device_fsa = {}
    end_device_ders = {}

    if config.cleanse_storage:
        adpt.clear_all_adapters()

    programs_by_description = {}

    for index, program_cfg in enumerate(config.programs):
        program_hrefs = hrefs.DERProgramHref(index)
        # Pop off default_der_control if specified.
        default_der_control = program_cfg.pop("DefaultDERControl", None)
        program = m.DERProgram(**program_cfg)
        program = program_hrefs.fill_hrefs(program)
        adpt.ListAdapter.append(hrefs.DEFAULT_DERP_ROOT, program)

        # Either set up default control or use the one passed in.
        if not default_der_control:
            default_der_control = m.DefaultDERControl(href=program_hrefs.default_control_href,
                                                      DERControlBase=m.DERControlBase())
        elif default_der_control:
            der_control_base = None
            if "DERControlBase" in default_der_control:
                der_control_base = default_der_control.pop("DERControlBase")
            default_der_control = m.DefaultDERControl(href=program.DefaultDERControlLink.href,
                                                      **default_der_control)
            if not der_control_base:
                default_der_control.DERControlBase = m.DERControlBase()
            else:
                default_der_control.DERControlBase = m.DERControlBase(**der_control_base)
        adpt.ListAdapter.initialize_uri(program.DERControlListLink.href, m.DERControl)
        add_href(program.DefaultDERControlLink.href, default_der_control)
        add_href(program.ActiveDERControlListLink.href, m.DERControlList(DERControl=[]))

        programs_by_description[program.description] = program

    fsa_with_description = {}

    for index, fsa in enumerate(config.fsas):
        programs = fsa.pop("programs", [])
        fsa_obj = m.FunctionSetAssignments(href=hrefs.fsa_href(index), **fsa)
        der_program_link = hrefs.SEP.join([fsa_obj.href, hrefs.DER_PROGRAM])
        fsa_obj.DERProgramListLink = m.DERProgramListLink(href=der_program_link, all=len(programs))
        adpt.ListAdapter.initialize_uri(der_program_link, m.DERProgram)
        for program in programs:
            if program not in programs_by_description:
                raise ValueError(
                    f"Program {program} not found in programs list for fsa {fsa['description']}")
            adpt.ListAdapter.append(der_program_link, programs_by_description[program])
        fsa_with_description[fsa_obj.description] = fsa_obj
        # put the programs back in the config fsa.
        fsa['programs'] = programs

    # Add DERCurves to the ListAdapter under the key hrefs.DEFAULT_CURVE_ROOT.
    for index, curve_cfg in enumerate(config.curves):
        curve = m.DERCurve(href=hrefs.SEP.join([hrefs.DEFAULT_CURVE_ROOT,
                                                str(index)]),
                           **curve_cfg)
        adpt.ListAdapter.append(hrefs.DEFAULT_CURVE_ROOT, curve)

    der_global_count = 0
    for index, cfg_device in enumerate(config.devices):

        create_device_capability(index)
        ed_href = hrefs.EndDeviceHref(index)
        end_device = adpt.EndDeviceAdapter.fetch_by_href(str(ed_href))
        if end_device is not None:
            _log.warning(
                f"End device {cfg_device.id} already exists.  Updating lfdi, sfdi, and postRate.")
            end_device.lFDI = tlsrepo.lfdi(cfg_device.id)
            end_device.sFDI = tlsrepo.sfdi(cfg_device.id)
            end_device.postRate = cfg_device.post_rate
            adpt.EndDeviceAdapter.put(index, end_device)
        else:
            _log.debug(f"Adding end device {cfg_device.id} to server")
            end_device = m.EndDevice(lFDI=tlsrepo.lfdi(cfg_device.id),
                                     sFDI=tlsrepo.sfdi(cfg_device.id),
                                     postRate=cfg_device.post_rate,
                                     enabled=True,
                                     changedTime=adpt.TimeAdapter.current_tick)
            add_enddevice(end_device)
            reg = m.Registration(href=end_device.RegistrationLink.href,
                                 pIN=cfg_device.pin,
                                 pollRate=cfg_device.poll_rate,
                                 dateTimeRegistered=adpt.TimeAdapter.current_tick)
            adpt.RegistrationAdapter.add(reg)
            add_href(reg.href, reg)

            if cfg_device.fsas:
                end_device.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(
                    href=ed_href.function_set_assignments, all=len(cfg_device.fsas))
                for fsa in cfg_device.fsas:
                    adpt.ListAdapter.append(ed_href.function_set_assignments,
                                            fsa_with_description[fsa])

            # If no ders specified then we create one for the end device.
            if cfg_device.ders:
                # Create references from the main der list to the ed specific list.
                for der in cfg_device.ders:
                    der_href = hrefs.DERHref(
                        hrefs.SEP.join([hrefs.DEFAULT_DER_ROOT,
                                        str(der_global_count)]))
                    der_global_count += 1
                    der_obj = m.DER(href=der_href.root,
                                    DERStatusLink=m.DERStatusLink(der_href.der_status),
                                    DERSettingsLink=m.DERSettingsLink(der_href.der_settings),
                                    DERCapabilityLink=m.DERCapabilityLink(der_href.der_capability),
                                    DERAvailabilityLink=m.DERAvailabilityLink(
                                        der_href.der_availability))
                    adpt.ListAdapter.append(ed_href.der_list, der_obj)
                    current_min_primacy = 10000
                    current_der_program = None
                    for fsa in adpt.ListAdapter.get_list(ed_href.function_set_assignments):
                        for der_program in adpt.ListAdapter.get_list(fsa.DERProgramListLink.href):
                            if current_der_program is None:
                                current_der_program = der_program
                            if der_program.primacy < current_min_primacy:
                                current_min_primacy = der_program.primacy
                    der_obj.CurrentDERProgramLink = m.CurrentDERProgramLink(
                        current_der_program.href)

            else:

                # der_href will manage the url links to other lists/resources for the DER.
                der_href = hrefs.DERHref(ed_href.der_list)

                # Create a reference to the default der list. Add an entry for the end device as
                # a DER object.  Note these are all available for the client to read/write via
                # GET/PUT to/from the server.
                der_list = m.DERList(DER=[
                    m.DER(href=der_href.root,
                          DERStatusLink=m.DERStatusLink(der_href.der_status),
                          DERSettingsLink=m.DERSettingsLink(der_href.der_settings),
                          DERCapabilityLink=m.DERCapabilityLink(der_href.der_capability),
                          DERAvailabilityLink=m.DERAvailabilityLink(der_href.der_availability))
                ])
                adpt.ListAdapter.append(ed_href.der_list, der_list)
                current_min_primacy = 10000
                current_der_program = None
                for fsa in adpt.ListAdapter.get_list(ed_href.function_set_assignments):
                    for der_program in adpt.ListAdapter.get_list(fsa.DERProgramListLink.href):
                        if current_der_program is None:
                            current_der_program = der_program
                        if der_program.primacy < current_min_primacy:
                            current_min_primacy = der_program.primacy
                der_list.DER[0].CurrentDERProgramLink = m.CurrentDERProgramLink(
                    current_der_program.href)

    adpt.ListAdapter.print_all()



import os

gridappsd_user = os.environ.get("GRIDAPPSD_USER", "system")
gridappsd_password = os.environ.get("GRIDAPPSD_PASSWORD", "manager")
gridappsd_address = os.environ.get("GRIDAPPSD_ADDRESS", "gridappsd")
gridappsd_port = os.environ.get("GRIDAPPSD_PORT", "61613")

os.environ['GRIDAPPSD_USER'] = gridappsd_user
os.environ['GRIDAPPSD_PASSWORD'] = gridappsd_password
os.environ['GRIDAPPSD_ADDRESS'] = gridappsd_address
os.environ['GRIDAPPSD_PORT'] = gridappsd_port
